import textwrap
from pathlib import Path

from popari.model import load_trained_model


def generate_mlproject_file(project_name: str, output_filepath: str = "MLproject"):
    """Generate MLproject file for command line use.

    Args:
        project_name: name of project

    """

    mlproject_contents = f"""
    name: {project_name}

    entry_points:
      train_debug:
        parameters:
          dataset_path: {{type: str, default: "input.h5ad"}}
          output_path: {{type: str, default: "output.h5ad"}}
          K: {{type: int}}
          dtype: {{type: str}}
          torch_device: {{type: str}}
          initial_device: {{type: str}}
          lambda_Sigma_x_inv: {{type: float, default: 1e-4}}
          lambda_Sigma_bar: {{type: float, default: 0}}
          spatial_affinity_groups: {{type: str, default: 'null'}}
          spatial_affinity_mode: {{type: str, default: 'shared lookup'}}
          metagene_groups: {{type: str, default: 'null'}}
          downsampling_method: {{type: str, default: 'partition'}}
          hierarchical_levels: {{type: int}}
          binning_downsample_rate: {{type: float, default: 0.2}}
          num_iterations: {{type: int}}
          spatial_preiterations: {{type: int}}
          nmf_preiterations: {{type: int}}
          verbose: {{type: int, default: 0}}
          random_state: {{type: int, default: 0}}
        command: "popari-mlflow --dataset_path={{dataset_path}} --output_path={{output_path}}
                                     --K={{K}} --dtype={{dtype}} --torch_device={{torch_device}}
                                     --initial_device={{initial_device}}
                                     --lambda_Sigma_x_inv={{lambda_Sigma_x_inv}}
                                     --lambda_Sigma_bar={{lambda_Sigma_bar}}
                                     --spatial_affinity_groups={{spatial_affinity_groups}}
                                     --spatial_affinity_mode={{spatial_affinity_mode}}
                                     --metagene_groups={{metagene_groups}}
                                     --downsampling_method={{downsampling_method}}
                                     --hierarchical_levels={{hierarchical_levels}}
                                     --binning_downsample_rate={{binning_downsample_rate}}
                                     --num_iterations={{num_iterations}}
                                     --spatial_preiterations={{spatial_preiterations}}
                                     --nmf_preiterations={{nmf_preiterations}}
                                     --verbose={{verbose}}
                                     --random_state={{random_state}}"

      train:
        parameters:
          dataset_path: {{type: str, default: "input.h5ad"}}
          output_path: {{type: str, default: "output.h5ad"}}
          K: {{type: int}}
          dtype: {{type: str}}
          torch_device: {{type: str}}
          initial_device: {{type: str}}
          lambda_Sigma_x_inv: {{type: float, default: 1e-4}}
          lambda_Sigma_bar: {{type: float, default: 0}}
          spatial_affinity_groups: {{type: str, default: 'null'}}
          spatial_affinity_mode: {{type: str, default: 'shared lookup'}}
          metagene_groups: {{type: str, default: 'null'}}
          downsampling_method: {{type: str, default: 'partition'}}
          hierarchical_levels: {{type: int}}
          binning_downsample_rate: {{type: float, default: 0.2}}
          num_iterations: {{type: int}}
          spatial_preiterations: {{type: int}}
          nmf_preiterations: {{type: int}}
          verbose: {{type: int, default: 0}}
          random_state: {{type: int, default: 0}}
        command: "popari-mlflow --dataset_path={{dataset_path}} --output_path={{output_path}}
                                     --K={{K}} --dtype={{dtype}} --torch_device={{torch_device}}
                                     --initial_device={{initial_device}}
                                     --lambda_Sigma_x_inv={{lambda_Sigma_x_inv}}
                                     --lambda_Sigma_bar={{lambda_Sigma_bar}}
                                     --spatial_affinity_groups={{spatial_affinity_groups}}
                                     --spatial_affinity_mode={{spatial_affinity_mode}}
                                     --metagene_groups={{metagene_groups}}
                                     --downsampling_method={{downsampling_method}}
                                     --hierarchical_levels={{hierarchical_levels}}
                                     --binning_downsample_rate={{binning_downsample_rate}}
                                     --num_iterations={{num_iterations}}
                                     --spatial_preiterations={{spatial_preiterations}}
                                     --nmf_preiterations={{nmf_preiterations}}
                                     --verbose={{verbose}}
                                     --random_state={{random_state}} > output_{{torch_device}}.txt 2>&1"

      popari:
        parameters:
          configuration_filepath: path
        command: "popari-grid-search --configuration_filepath {{configuration_filepath}}"
    """

    with open(output_filepath, "w") as f:
        f.writelines(textwrap.dedent(mlproject_contents).strip())


def load_from_mlflow_run(run, client, file_suffix="_result", **loading_kwargs):
    """Load trained model from MLflow run.

    Args:
        run: MLflow run containing model artifact.
        client: MLflow client initialized to point to experiment location.
        loading_kwargs: kwargs that can be passed to Popari pretrained initialization.

    """
    run_id = run.info.run_id
    artifacts = client.list_artifacts(run_id)

    (trained_model_artifact,) = list(filter(lambda artifact: Path(artifact.path).stem.endswith(file_suffix), artifacts))

    print(run.data.params)
    _, base_uri = run.info.artifact_uri.split(":")
    trained_model_filepath = Path(base_uri) / trained_model_artifact.path

    trained_model = load_trained_model(trained_model_filepath, **loading_kwargs)

    return trained_model



from pathlib import Path
from typing import Optional, Sequence, Union

import anndata as ad
import awkward as ak
import numpy as np
import torch
from scipy.sparse import csr_array, issparse

from popari.components import PopariDataset
from popari.util import concatenate, convert_adjacency_matrix_to_awkward_array, unconcatenate


def load_anndata(filepath: Union[str, Path]):
    """Load AnnData object from h5ad file and reformat for Popari."""

    merged_dataset = ad.read_h5ad(filepath)

    datasets, replicate_names = unmerge_anndata(merged_dataset)

    return datasets, replicate_names


def unmerge_anndata(merged_dataset: ad.AnnData):
    """Unmerge composite AnnData object into constituent datasets."""

    merged_dataset.X = csr_array(merged_dataset.X)
    datasets = unconcatenate(merged_dataset)

    for dataset in datasets:
        replicate_string = f"{dataset.name}"
        if "Sigma_x_inv" in dataset.uns:
            # Keep only Sigma_x_inv corresponding to a particular replicate
            replicate_Sigma_x_inv = dataset.uns["Sigma_x_inv"][replicate_string]
            if np.isscalar(replicate_Sigma_x_inv) and replicate_Sigma_x_inv == -1:
                replicate_Sigma_x_inv = None

            dataset.uns["Sigma_x_inv"] = {
                replicate_string: replicate_Sigma_x_inv,
            }

        if "Sigma_x_inv_bar" in dataset.uns:
            # Keep only Sigma_x_inv_bar corresponding to a particular replicate
            replicate_Sigma_x_inv_bar = dataset.uns["Sigma_x_inv_bar"][replicate_string]
            if np.isscalar(replicate_Sigma_x_inv_bar) and replicate_Sigma_x_inv_bar == -1:
                replicate_Sigma_x_inv_bar = None

            dataset.uns["Sigma_x_inv_bar"] = {
                replicate_string: replicate_Sigma_x_inv_bar,
            }

        # Hacks to load adjacency matrices efficiently
        if "adjacency_matrix" in dataset.uns:
            dataset.obsp["adjacency_matrix"] = csr_array(
                dataset.uns["adjacency_matrix"][replicate_string],
            )

        adjacency_matrix = dataset.obsp["adjacency_matrix"].tocoo()

        if "adjacency_list" not in dataset.obsm:
            dataset.obsm["adjacency_list"] = convert_adjacency_matrix_to_awkward_array(adjacency_matrix)

        if "M" in dataset.uns:
            if replicate_string in dataset.uns["M"]:
                replicate_M = make_hdf5_compatible(dataset.uns["M"][replicate_string])
                dataset.uns["M"] = {replicate_string: replicate_M}

        if "M_bar" in dataset.uns:
            if replicate_string in dataset.uns["M_bar"]:
                replicate_M_bar = make_hdf5_compatible(
                    dataset.uns["M_bar"][replicate_string],
                )
                dataset.uns["M_bar"] = {replicate_string: replicate_M_bar}

        if "popari_hyperparameters" in dataset.uns:
            if "prior_x" in dataset.uns["popari_hyperparameters"]:
                prior_x = make_hdf5_compatible(
                    dataset.uns["popari_hyperparameters"]["prior_x"],
                )
                dataset.uns["popari_hyperparameters"]["prior_x"] = prior_x

            if "spatial_affinity_groups" in dataset.uns["popari_hyperparameters"]:
                name_parts = dataset.name.split("_level_")
                if len(name_parts) > 1:
                    *_, level = name_parts
                    level = int(level)
                else:
                    level = 0

                groups = dataset.uns["popari_hyperparameters"]["spatial_affinity_groups"]
                filtered_groups = {}
                for group, group_replicates in groups.items():
                    group_name_parts = group.split("_level_")
                    if len(group_name_parts) > 1:
                        *_, group_level = group_name_parts
                        group_level = int(group_level)
                    else:
                        group_level = 0

                    if group_level == level:
                        filtered_groups[group] = group_replicates

                dataset.uns["popari_hyperparameters"]["spatial_affinity_groups"] = filtered_groups

            if "metagene_groups" in dataset.uns["popari_hyperparameters"]:
                name_parts = dataset.name.split("_level_")
                if len(name_parts) > 1:
                    *_, level = name_parts
                    level = int(level)
                else:
                    level = 0

                groups = dataset.uns["popari_hyperparameters"]["metagene_groups"]
                filtered_groups = {}
                for group, group_replicates in groups.items():
                    group_name_parts = group.split("_level_")
                    if len(group_name_parts) > 1:
                        *_, group_level = group_name_parts
                        group_level = int(group_level)
                    else:
                        group_level = 0

                    if group_level == level:
                        filtered_groups[group] = group_replicates

                dataset.uns["popari_hyperparameters"]["metagene_groups"] = filtered_groups

        # if "X" in dataset.obsm:
        #     replicate_X = make_hdf5_compatible(dataset.obsm["X"])
        #     dataset.obsm["X"] = replicate_X

    replicate_names = [dataset.name for dataset in datasets]
    return datasets, replicate_names


def merge_anndata(datasets: Sequence[PopariDataset], ignore_raw_data: bool = False):
    """Merge multiple PopariDatasets into a single AnnData object (for
    storage)."""

    dataset_copies = []
    for dataset in datasets:
        replicate = dataset.name
        replicate_string = f"{replicate}"
        if ignore_raw_data:
            dataset.X = csr_array(dataset.X.shape)
        else:
            dataset.X = csr_array(dataset.X)

        dataset_copy = PopariDataset(dataset, dataset.name)

        # Hacks to store adjacency matrices efficiently
        if "adjacency_matrix" in dataset.obsp:
            adjacency_matrix = make_hdf5_compatible(dataset.obsp["adjacency_matrix"])
            dataset_copy.uns["adjacency_matrix"] = {replicate_string: adjacency_matrix}
        elif "adjacency_matrix" in dataset.uns:
            dataset_copy.uns["adjacency_matrix"] = {
                replicate_string: make_hdf5_compatible(adjacency_matrix)
                for replicate_string, adjacency_matrix in dataset.uns["adjacency_matrix"]
            }

        if "Sigma_x_inv" in dataset_copy.uns:
            replicate_Sigma_x_inv = dataset_copy.uns["Sigma_x_inv"][replicate_string]
            if replicate_string in dataset_copy.uns["Sigma_x_inv"]:
                # Using a sentinel value - hopefully this can be fixed in the future!
                if replicate_Sigma_x_inv is None:
                    replicate_Sigma_x_inv = -1
                else:
                    replicate_Sigma_x_inv = make_hdf5_compatible(replicate_Sigma_x_inv)
            dataset_copy.uns["Sigma_x_inv"] = {replicate_string: replicate_Sigma_x_inv}

        if "Sigma_x_inv_bar" in dataset_copy.uns:
            replicate_Sigma_x_inv_bar = dataset_copy.uns["Sigma_x_inv_bar"][replicate_string]
            if replicate_string in dataset_copy.uns["Sigma_x_inv_bar"]:
                # Using a sentinel value - hopefully this can be fixed in the future!
                if replicate_Sigma_x_inv_bar is None:
                    replicate_Sigma_x_inv_bar = -1
                else:
                    replicate_Sigma_x_inv_bar = make_hdf5_compatible(
                        replicate_Sigma_x_inv_bar,
                    )
            dataset_copy.uns["Sigma_x_inv_bar"] = {
                replicate_string: replicate_Sigma_x_inv_bar,
            }

        if "M" in dataset_copy.uns:
            if replicate_string in dataset_copy.uns["M"]:
                replicate_M = make_hdf5_compatible(
                    dataset_copy.uns["M"][replicate_string],
                )
                dataset_copy.uns["M"] = {replicate_string: replicate_M}

        if "M_bar" in dataset_copy.uns:
            if replicate_string in dataset_copy.uns["M_bar"]:
                replicate_M_bar = make_hdf5_compatible(
                    dataset_copy.uns["M_bar"][replicate_string],
                )
                dataset_copy.uns["M_bar"] = {replicate_string: replicate_M_bar}

        if "popari_hyperparameters" in dataset_copy.uns:
            if "prior_x" in dataset_copy.uns["popari_hyperparameters"]:
                prior_x = make_hdf5_compatible(
                    dataset_copy.uns["popari_hyperparameters"]["prior_x"],
                )
                dataset_copy.uns["popari_hyperparameters"]["prior_x"] = prior_x

        if "dataset_name" not in dataset_copy.uns:
            dataset_copy.uns["dataset_name"] = dataset.name

        # if "X" in dataset_copy.obsm:
        #     replicate_X = make_hdf5_compatible(dataset_copy.obsm["X"])
        #     dataset_copy.obsm["X"] = replicate_X
        dataset_copies.append(dataset_copy)

    merged_dataset = concatenate(dataset_copies, join="outer")

    return merged_dataset


def save_anndata(
    filepath: Union[str, Path],
    datasets: Sequence[PopariDataset],
    ignore_raw_data: bool = False,
):
    """Save Popari state as AnnData object."""

    merged_dataset = merge_anndata(datasets, ignore_raw_data=ignore_raw_data)
    merged_dataset.write(filepath)

    return merged_dataset


def make_hdf5_compatible(array: Union[torch.Tensor, np.ndarray]):
    """Convert tensors to Numpy array for compatibility with HDF5.

    Args:
        array: array to convert to Numpy.

    """
    if type(array) == torch.Tensor:
        cpu_tensor = array.detach().cpu()
        if cpu_tensor.is_sparse:
            cpu_tensor = cpu_tensor.to_dense()
        return cpu_tensor.numpy()
    return array



from copy import deepcopy
from typing import Sequence

import anndata as ad
import awkward as ak
import numpy as np
import pandas as pd
import scanpy as sc
import seaborn as sns
import squidpy as sq
from scipy.sparse import csr_matrix
from tqdm.auto import tqdm, trange


class PopariDataset(ad.AnnData):
    r"""Wrapper around AnnData object. Allows for preprocessing of dataset for
    Popari.

    Attributes:
        dataset: AnnData dataset to convert into Popari-compatible object.
        replicate_name: name of dataset
        coordinates_key: location in ``.obsm`` dataframe of 2D coordinates for datapoints.

    """

    def __init__(
        self,
        dataset: ad.AnnData,
        replicate_name: str,
        coordinates_key: str = "spatial",
    ):
        super().__init__(
            X=dataset.X,
            obs=dataset.obs,
            obsm=dataset.obsm,
            obsp=dataset.obsp,
            var=dataset.var,
            varm=dataset.varm,
            varp=dataset.varp,
            uns=deepcopy(dataset.uns),
        )

        self.coordinates_key = coordinates_key
        if self.coordinates_key not in self.obsm:
            raise ValueError("Dataset must include spatial coordinates.")

        self.name = f"{replicate_name}"

    def compute_spatial_neighbors(self, threshold: float = 94.5):
        r"""Compute neighbor graph based on spatial coordinates.

        Stores resulting graph in ``self.obs["adjacency_list"]``.

        """

        sq.gr.spatial_neighbors(self, coord_type="generic", delaunay=True)
        distance_matrix = self.obsp["spatial_distances"]
        distances = distance_matrix.data
        cutoff = np.percentile(distances, threshold)

        sq.gr.spatial_neighbors(
            self,
            coord_type="generic",
            delaunay=True,
            radius=[0, cutoff],
        )
        self.obsp["adjacency_matrix"] = self.obsp["spatial_connectivities"]

        num_cells, _ = self.obsp["adjacency_matrix"].shape

        adjacency_list = [[] for _ in range(num_cells)]
        for x, y in zip(*self.obsp["adjacency_matrix"].nonzero()):
            adjacency_list[x].append(y)

        self.obsm["adjacency_list"] = ak.Array(adjacency_list)

    @staticmethod
    def remove_connectivity_artifacts(
        sparse_distance_matrix: csr_matrix,
        sparse_adjacency_matrix: csr_matrix,
        threshold: float = 94.5,
    ):
        """Remove artifacts in adjacency matrices produced by heuristic
        algorithms.

        For example, when Delaunay triangulation is used to produce the adjacency matrix, some spots which are
        connected as a result may actually be far apart in Euclidean space.

        Args:
            sparse_distance_matrix:

        """
        dense_distances = sparse_distance_matrix.toarray()
        distances = sparse_distance_matrix.data
        cutoff = np.percentile(distances, threshold)
        mask = dense_distances < cutoff

        sparse_adjacency_matrix[~mask] = 0
        sparse_adjacency_matrix.eliminate_zeros()

        return sparse_adjacency_matrix

    def plot_metagene_embedding(self, metagene_index: int, embedding_key: str = "X", **scatterplot_kwargs):
        r"""Plot the embedding values for a embedding in-situ.

        Args:
            metagene_index: index of the embedding to plot.
            **scatterplot_kwargs: keyword args to pass to ``sns.scatterplot``.

        """
        points = self.obsm["spatial"]
        x, y = points.T
        embedding = self.obsm[embedding_key][:, metagene_index]

        biased_batch_effect = pd.DataFrame({"x": x, "y": y, f"Metagene {metagene_index}": embedding})
        sns.scatterplot(data=biased_batch_effect, x="x", y="y", hue=f"Metagene {metagene_index}", **scatterplot_kwargs)



from functools import partial, wraps
from typing import Callable, Mapping, Optional, Sequence

import anndata as ad
import matplotlib.patches as patches
import mpl_toolkits.axisartist.angle_helper as angle_helper
import mpl_toolkits.axisartist.floating_axes as floating_axes
import networkx as nx
import numpy as np
import pandas as pd
import scanpy as sc
import scanpy.external as sce
import seaborn as sns
import squidpy as sq
from matplotlib import cm, colormaps
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from matplotlib.colors import ListedColormap
from matplotlib.transforms import Affine2D
from mpl_toolkits.axisartist.grid_finder import DictFormatter, FixedLocator, MaxNLocator
from scipy.sparse import csr_array, csr_matrix, issparse
from scipy.stats import wilcoxon, zscore
from sklearn.metrics import adjusted_rand_score, confusion_matrix, precision_score, silhouette_score
from sklearn.model_selection import train_test_split
from sklearn.neighbors import KNeighborsClassifier
from sklearn.preprocessing import LabelEncoder

from popari._binning_utils import chunked_downsample_on_grid, filter_gridpoints
from popari._popari_dataset import PopariDataset
from popari.util import (
    compute_neighborhood_enrichment,
    concatenate,
    get_metagene_signature,
    normalize_expression_by_threshold,
    run_gsea,
    smooth_labels,
    smooth_metagene_expression,
    unconcatenate,
)


def setup_squarish_axes(num_axes, **subplots_kwargs):
    """Create matplotlib subplots as squarely as possible."""

    height = int(np.sqrt(num_axes))
    width = num_axes // height
    height += width * height != num_axes

    constrained_layout = (
        True if "constrained_layout" not in subplots_kwargs else subplots_kwargs.pop("constrained_layout")
    )
    dpi = subplots_kwargs.pop("dpi", 300)
    sharex = subplots_kwargs.pop("sharex", True)
    sharey = subplots_kwargs.pop("sharey", True)

    fig, axes = plt.subplots(
        height,
        width,
        squeeze=False,
        constrained_layout=constrained_layout,
        dpi=dpi,
        sharex=sharex,
        sharey=sharey,
        **subplots_kwargs,
    )

    return fig, axes


def copy_annotations(original_dataset, updated_dataset, annotations: Optional[Mapping[str, Sequence[str]]] = None):
    """Copy updates from one AnnData to another.

    Note that this is a bit wasteful, as it copies every key in a given
    annotation layer.

    """
    if annotations is None:
        annotations = dict()

    for annotation, keys in annotations.items():
        original_annotation = getattr(original_dataset, annotation)
        updated_annotation = getattr(updated_dataset, annotation)

        if keys is None:
            keys = updated_annotation.keys()

        for key in keys:
            original_annotation[key] = updated_annotation[key]


def for_model(function=None, *, return_outputs: bool = False):
    if function is None:
        return partial(for_model, return_outputs=return_outputs)

    @wraps(function)
    def model_wrapper(trained_model: "Popari", *args, **kwargs):
        level = kwargs.pop("level", 0)
        datasets = trained_model.hierarchy[level].datasets

        outputs = function(datasets, *args, **kwargs)

        return outputs if return_outputs else None

    return model_wrapper


def enable_joint(function=None, *, annotations: Optional[Mapping[str, Sequence[str]]] = None):
    """Decorator to extend functions that operate on lists of dataset to work on
    a joint, merged dataset.

    Args:
        function: function to decorate. Must accept a list of datasets as its first argument.

    """
    if function is None:
        return partial(enable_joint, annotations=annotations)

    @wraps(function)
    def joint_wrapper(datasets, *args, **kwargs):
        """
        Args:
            joint: if `True`, apply `function` to merged dataset.
        """
        joint = kwargs.pop("joint", False)
        if joint:
            original_datasets = datasets
            merged_dataset = concatenate(datasets)
            datasets = [merged_dataset]

        outputs = function(datasets, *args, **kwargs)

        if joint:
            unmerged_datasets = unconcatenate(merged_dataset)

            for unmerged_dataset, original_dataset in zip(unmerged_datasets, original_datasets):
                copy_annotations(original_dataset, unmerged_dataset, annotations=annotations)

            outputs = original_datasets, outputs

        return outputs

    return joint_wrapper


def broadcast(function):
    """Decorator to extend a function from a single dataset to multiple
    datasets.

    TODO: Doesn't work in the case that the function's return values are meaningful...

    """

    @wraps(function)
    def broadcast_wrapper(datasets, *args, **kwargs):
        """"""

        outputs = []
        for dataset in datasets:
            output = function(dataset, *args, **kwargs)
            outputs.append(output)

        if len(outputs) == 1:
            outputs = outputs[0]

        return outputs

    return broadcast_wrapper


def broadcast_plottable(function):
    """Decorator to extend a plottable function from a single dataset to
    multiple datasets.

    TODO: Doesn't work in the case that the function's return values are meaningful...

    """

    @wraps(function)
    def broadcast_wrapper(datasets, *args, **kwargs):
        """"""
        axes = kwargs.pop("axes", None)

        sharex = kwargs.pop("sharex", False)
        sharey = kwargs.pop("sharey", False)

        fig = None
        if axes is None:
            fig, axes = setup_squarish_axes(len(datasets), sharex=sharex, sharey=sharey)

        outputs = []
        for dataset, ax in zip(datasets, axes.flat):
            output = function(dataset, *args, ax=ax, **kwargs)
            outputs.append(output)

        if len(outputs) == 1:
            outputs = outputs[0]

        return fig, outputs

    return broadcast_wrapper


def _preprocess_embeddings(
    datasets: Sequence[PopariDataset],
    input_key="X",
    normalized_key="normalized_X",
):
    """Normalize embeddings per each cell.

    This step helps to make cell embeddings comparable, and facilitates
    downstream tasks like clustering.

    """
    for dataset in datasets:
        if "X" not in dataset.obsm:
            raise ValueError("Must initialize embeddings before normalizing them.")

        normalized_embeddings = zscore(dataset.obsm[input_key])
        nan_mask = np.isnan(normalized_embeddings)
        normalized_embeddings[nan_mask] = 0

        dataset.obsm[normalized_key] = normalized_embeddings
        sc.pp.neighbors(dataset, use_rep=normalized_key)

    return datasets


def _plot_metagene_embedding(
    datasets: Sequence[PopariDataset],
    metagene_index: int,
    axes: Optional[Sequence[Axes]] = None,
    **scatterplot_kwargs,
):
    r"""Plot a single metagene in-situ across all datasets.

    Args:
        datasets: list of datasets to plot
        metagene_index: the index of the metagene to plot.
        axes: A predefined set of matplotlib axes to plot on.

    """

    legend = scatterplot_kwargs.pop("legend", False)
    default_s = scatterplot_kwargs.pop("s", None)
    linewidth = scatterplot_kwargs.pop("linewidth", 0)
    palette = scatterplot_kwargs.pop("palette", "viridis")
    dpi = scatterplot_kwargs.pop("dpi", 100)

    if axes is None:
        fig, axes = setup_squarish_axes(len(datasets), sharex=False, sharey=False, dpi=dpi)

    else:
        fig = axes.flat[0].get_figure()

    for index in range(len(datasets), axes.size):
        axes.flat[index].axis("off")

    for dataset, ax in zip(datasets, axes.flat):
        if default_s is None:
            s = round(2000 / len(dataset))
        else:
            s = default_s

        ax.set_aspect("equal", "box")
        ax.invert_yaxis()
        ax.set_xticks([], [])  # note you need two lists one for the positions and one for the labels
        ax.set_yticks([], [])  # same for y ticks
        dataset.plot_metagene_embedding(
            metagene_index,
            legend=legend,
            s=s,
            linewidth=linewidth,
            palette=palette,
            ax=ax,
            **scatterplot_kwargs,
        )

    fig.suptitle(f"Metagene {metagene_index}")

    return fig


# TODO: we need to copy `.uns` to keep all added info. But this is buggy, because copying `.uns` is not done correctly
@enable_joint(annotations={"obs": None, "uns": ["leiden"], "obsp": None})
def _leiden(
    datasets: Sequence[PopariDataset],
    resolution: float = 1.0,
    tolerance: float = 0.05,
    **kwargs,
):
    r"""Compute Leiden clustering for all datasets.

    Args:
        trained_model: the trained Popari model.
        joint: if `True`, jointly cluster the spots
        datasets: list of datasets to cluster
        use_rep: the key in the ``.obsm`` dataframe to ue as input to the Leiden clustering algorithm.
        resolution: the resolution to use for Leiden clustering. Higher values yield finer clusters..

    """
    # n_iterations = kwargs.pop("n_iterations", 2) # TODO: add these parameters back in after resolving issue
    _cluster.__wrapped__(
        datasets,
        resolution=resolution,
        method="leiden",
        tolerance=tolerance,
        # flavor="igraph",
        # n_iterations=n_iterations,
        **kwargs,
    )


@enable_joint(annotations={"obs": None, "uns": None, "obsp": None})
def _cluster(
    datasets: Sequence[PopariDataset],
    use_rep="normalized_X",
    method: str = "leiden",
    n_neighbors: int = 20,
    target_clusters: Optional[int] = None,
    tolerance: float = 0.005,
    compute_neighbors: bool = True,
    verbose: bool = False,
    **kwargs,
):
    r"""Compute clustering for all datasets.

    Args:
        datasets: list of datasets to cluster
        use_rep: the key in the ``.obsm`` dataframe to ue as input to the Leiden clustering algorithm.
        resolution: the resolution to use for Leiden clustering. Higher values yield finer clusters.

    """

    clustering_function = getattr(sc.tl, method)

    random_state = kwargs.pop("random_state", 0)
    resolution = kwargs.pop("resolution", 1.0)
    key_added = kwargs.pop("key_added", method)

    for dataset in datasets:
        if compute_neighbors:
            sc.pp.neighbors(dataset, use_rep=use_rep, random_state=random_state, n_neighbors=n_neighbors)

        clustering_function(dataset, resolution=resolution, random_state=random_state, key_added=key_added, **kwargs)

        num_clusters = len(dataset.obs[key_added].unique())

        lower_bound = 0.1 * resolution
        upper_bound = 10 * resolution
        while target_clusters and num_clusters != target_clusters and np.abs(lower_bound - upper_bound) > tolerance:
            effective_resolution = (lower_bound * upper_bound) ** 0.5
            clustering_function(
                dataset,
                resolution=effective_resolution,
                random_state=random_state,
                key_added=key_added,
                **kwargs,
            )
            num_clusters = len(dataset.obs[key_added].unique())
            if num_clusters < target_clusters:
                lower_bound = effective_resolution
            elif num_clusters >= target_clusters:
                upper_bound = effective_resolution

            if verbose:
                print(f"Current number of clusters: {num_clusters}")
                print(f"Resolution: {effective_resolution}")

    return datasets


@enable_joint(annotations={"obsm": ["X_pca"], "varm": ["PCs"], "uns": ["pca"]})
@broadcast
def _pca(dataset: PopariDataset, n_comps: int = 50, **pca_kwargs):
    r"""Compute PCA for all datasets.

    Args:
        datasets: list of datasets to process

    """
    dataset.X = csr_matrix(dataset.X)  # Hack because sparse PCA isn't implemented
    sc.pp.pca(dataset, n_comps=n_comps, **pca_kwargs)
    dataset.X = csr_array(dataset.X)


@enable_joint(annotations={"obsm": ["X_umap"], "uns": ["umap"]})
@broadcast
def _umap(dataset: PopariDataset, use_rep: str = "X", compute_neighbors: bool = True, n_neighbors: int = 20):
    r"""Compute UMAP for all datasets.

    Args:
        datasets: list of datasets to process

    """

    if compute_neighbors:
        sc.pp.neighbors(dataset, use_rep=use_rep, n_neighbors=n_neighbors)

    sc.tl.umap(dataset)


@enable_joint
@broadcast
def _plot_in_situ(dataset: Sequence[PopariDataset], axes=None, fig=None, color="leiden", **spatial_kwargs):
    r"""Plot a categorical label across all datasets in-situ.

    Extends AnnData's ``sc.pl.spatial`` function to plot labels/values across multiple replicates.

    Args:
        datasets: list of datasets to process
        color: the key in the ``.obs`` dataframe to plot.
        axes: A predefined set of matplotlib axes to plot on.

    """

    sharex = False if "sharex" not in spatial_kwargs else spatial_kwargs.pop("sharex")
    sharey = False if "sharey" not in spatial_kwargs else spatial_kwargs.pop("sharey")
    dpi = spatial_kwargs.pop("dpi", 100)

    # fig = None
    if axes is None:
        fig, axes = setup_squarish_axes(1, sharex=sharex, sharey=sharey, dpi=dpi)

    edges_width = spatial_kwargs.pop("edges_width", 0.2)
    default_size = spatial_kwargs.pop("size", None)
    palette = spatial_kwargs.pop("palette", None)
    palette = sc.pl.palettes.godsnot_102 if not palette else plt.get_cmap(palette).colors
    palette = ListedColormap(palette)

    legend_fontsize = spatial_kwargs.pop("legend_fontsize", "xx-small")
    edgecolors = spatial_kwargs.pop("edgecolors", "none")
    connectivity_key = spatial_kwargs.pop("connectivity_key", "adjacency_matrix")
    shape = spatial_kwargs.pop("shape", None)

    neighbors_key = spatial_kwargs.pop("neighbors_key", "spatial_neighbors")

    size = 1000 / len(dataset)
    if default_size is not None:
        size *= default_size

    total_categories = len(dataset.obs[color].cat.categories)
    leiden_colorlist = np.array(palette.colors)[np.linspace(0, len(palette.colors) - 1, total_categories, dtype=int)]
    leiden_colors = dict(zip([str(category) for category in dataset.obs[color].cat.categories], leiden_colorlist))
    datasets = unconcatenate(dataset)

    for dataset, ax in zip(datasets, axes.flat):
        dataset_categories = dataset.obs[color].cat.categories
        palette = ListedColormap([color for category, color in leiden_colors.items() if category in dataset_categories])
        sq.pl.spatial_scatter(
            dataset,
            shape=shape,
            size=size,
            connectivity_key=connectivity_key,
            color=color,
            edges_width=edges_width,
            legend_fontsize=legend_fontsize,
            ax=ax,
            fig=fig,
            palette=palette,
            edgecolors=edgecolors,
            library_key="batch",
            **spatial_kwargs,
        )

    return fig


@enable_joint
@broadcast_plottable
def _plot_umap(dataset: PopariDataset, color="leiden", ax=None, **kwargs):
    r"""Plot a categorical label across all datasets in-situ.

    Extends AnnData's ``sc.pl.spatial`` function to plot labels/values across multiple replicates.

    Args:
        datasets: list of datasets to process
        color: the key in the ``.obs`` dataframe to plot.
        axes: A predefined set of matplotlib axes to plot on.

    """

    edges_width = kwargs.pop("edges_width", 0.2)
    size = kwargs.pop("size", 0.04)
    edges = kwargs.pop("edges", sc.pl.palettes.godsnot_102)
    palette = kwargs.pop("palette", sc.pl.palettes.godsnot_102)
    legend_fontsize = kwargs.pop("legend_fontsize", "xx-small")
    neighbors_key = kwargs.pop("neighbors_key", "neighbors")

    sc.pl.umap(
        dataset,
        size=size,
        neighbors_key=neighbors_key,
        color=color,
        edges=edges,
        edges_width=edges_width,
        legend_fontsize=legend_fontsize,
        ax=ax,
        show=False,
        palette=palette,
        **kwargs,
    )


def _multireplicate_heatmap(
    datasets: Sequence[PopariDataset],
    title_font_size: Optional[int] = None,
    axes: Optional[Sequence[Axes]] = None,
    obsm: Optional[str] = None,
    obsp: Optional[str] = None,
    uns: Optional[str] = None,
    label_values: bool = False,
    label_font_size: int = None,
    nested: bool = True,
    mask: Optional[np.ndarray] = None,
    **heatmap_kwargs,
):
    r"""Plot 2D heatmap data across all datasets.

    Wrapper function to enable plotting of continuous 2D data across multiple replicates. Only
    one of ``obsm``, ``obsp`` or ``uns`` should be used.
    Args:
        datasets: list of datasets to process
        axes: A predefined set of matplotlib axes to plot on.
        obsm: the key in the ``.obsm`` dataframe to plot.
        obsp: the key in the ``.obsp`` dataframe to plot.
        uns: the key in the ``.uns`` dataframe to plot. Unstructured data must be 2D in shape.
        **heatmap_kwargs: arguments to pass to the `ax.imshow` call for each dataset

    """

    sharex = heatmap_kwargs.pop("sharex", True)
    sharey = heatmap_kwargs.pop("sharey", True)
    dpi = heatmap_kwargs.pop("dpi", 100)

    fig = None
    if axes is None:
        fig, axes = setup_squarish_axes(len(datasets), sharex=sharex, sharey=sharey, dpi=dpi)

    aspect = heatmap_kwargs.pop("aspect", 1)
    cmap = heatmap_kwargs.pop("cmap", "hot")

    images = []
    for dataset_index, ax in enumerate(axes.flat):
        if dataset_index >= len(datasets):
            ax.set_visible(False)
            continue

        dataset = datasets[dataset_index]
        key = None
        if obsm:
            image = dataset.obsm[obsm]
        if obsp:
            image = dataset.obsp[obsp]
        if uns:
            image = dataset.uns[uns]

        if nested:
            image = image[dataset.name]

        if mask is not None:
            image = np.ma.masked_where(mask, image)

        images.append(image)

    for dataset_index, (image, ax) in enumerate(zip(images, axes.flat)):
        dataset = datasets[dataset_index]
        im = ax.imshow(image, cmap=cmap, interpolation="nearest", aspect=aspect, **heatmap_kwargs)
        if title_font_size is not None:
            ax.set_title(dataset.name, fontsize=title_font_size)
        else:
            ax.set_title(dataset.name)

        if label_values:
            truncated_image = image.astype(int)
            for (j, i), label in np.ndenumerate(truncated_image):
                if mask is None or not mask[j, i]:
                    ax.text(i, j, label, ha="center", va="center", fontsize=label_font_size)

        num_rows, num_columns = image.shape
        ax.set_xticks(np.arange(num_columns).astype(int))
        ax.set_yticks(np.arange(num_rows).astype(int))

        plt.colorbar(im, ax=ax, orientation="vertical", fraction=0.046, pad=0.04)

    return fig


def _multigroup_heatmap(
    datasets: Sequence[PopariDataset],
    groups: dict,
    title_font_size: Optional[int] = None,
    axes: Optional[Sequence[Axes]] = None,
    key: Optional[str] = None,
    label_values: bool = False,
    label_font_size: int = None,
    **heatmap_kwargs,
):
    r"""Plot 2D heatmap data across all datasets.

    Wrapper function to enable plotting of continuous 2D data across multiple replicates. Only
    one of ``obsm``, ``obsp`` or ``uns`` should be used.

    Args:
        datasets: list of datasets to process
        axes: A predefined set of matplotlib axes to plot on.
        obsm: the key in the ``.obsm`` dataframe to plot.
        obsp: the key in the ``.obsp`` dataframe to plot.
        uns: the key in the ``.uns`` dataframe to plot. Unstructured data must be 2D in shape.
        **heatmap_kwargs: arguments to pass to the `ax.imshow` call for each dataset

    """

    sharex = True if "sharex" not in heatmap_kwargs else heatmap_kwargs.pop("sharex", True)
    sharey = True if "sharey" not in heatmap_kwargs else heatmap_kwargs.pop("sharey", True)

    fig = None
    if axes is None:
        fig, axes = setup_squarish_axes(len(groups), sharex=sharex, sharey=sharey)

    aspect = heatmap_kwargs.pop("aspect", 0.05)
    cmap = heatmap_kwargs.pop("cmap", "hot")

    for group_index, (ax, group_name) in enumerate(zip(axes.flat, groups)):
        first_dataset_name = groups[group_name][0]
        first_dataset = next(filter(lambda dataset: dataset.name == first_dataset_name, datasets))

        if group_index > len(groups):
            ax.set_visible(False)
            continue

        image = first_dataset.uns[key][group_name]

        im = ax.imshow(image, cmap=cmap, interpolation="nearest", aspect=aspect, **heatmap_kwargs)
        if title_font_size is not None:
            ax.set_title(group_name, fontsize=title_font_size)
        if label_values:
            truncated_image = image.astype(int)
            for (j, i), label in np.ndenumerate(truncated_image):
                ax.text(i, j, label, ha="center", va="center", fontsize=label_font_size)

        fig.colorbar(im, ax=ax, orientation="vertical")

    return fig


def _compute_empirical_correlations(
    datasets: Sequence[PopariDataset],
    scaling: float,
    feature: str = "X",
    output: str = "empirical_correlation",
):
    """Compute the empirical spatial correlation for a feature set across all
    datasets.

    Args:
        datasets: list of datasets to process
        feature: key in `.obsm` of feature set for which spatial correlation should be computed.
        output: key in `.uns` where output correlation matrices should be stored.

    """

    num_replicates = len(datasets)

    first_dataset = datasets[0]
    _, K = first_dataset.obsm[feature].shape
    empirical_correlations = np.zeros([num_replicates, K, K])
    for replicate, dataset in enumerate(datasets):
        adjacency_list = dataset.obsm["adjacency_list"]
        X = dataset.obsm[feature]
        Z = X / np.linalg.norm(X, axis=1, keepdims=True, ord=1)
        edges = np.array([(i, j) for i, e in enumerate(adjacency_list) for j in e])

        x = Z[edges[:, 0]]
        y = Z[edges[:, 1]]
        x = x - x.mean(axis=0, keepdims=True)
        y = y - y.mean(axis=0, keepdims=True)
        y_std = y.std(axis=0, keepdims=True)
        x_std = x.std(axis=0, keepdims=True)
        corr = (y / y_std).T @ (x / x_std) / len(x)
        empirical_correlations[replicate] = -corr

    # Convert nan values to 0
    empirical_correlations = np.nan_to_num(empirical_correlations)

    # Symmetrizing and zero-centering empirical_correlation
    empirical_correlations = (empirical_correlations + np.transpose(empirical_correlations, (0, 2, 1))) / 2
    empirical_correlations -= empirical_correlations.mean(axis=(1, 2), keepdims=True)
    empirical_correlations *= scaling

    for dataset, empirical_correlation in zip(datasets, empirical_correlations):
        all_correlations = {dataset.name: empirical_correlation}
        dataset.uns[output] = all_correlations

    return datasets


def _adjacency_permutation_test(
    dataset: PopariDataset,
    labels: str = "X",
    n_trials: int = 500,
    random_state: int = 0,
    pvalue_key: str = "pvalue",
):
    r"""Compute p-values for neighborhood enrichment.

    See
    `In silico tissue generation and power analysis for spatial omics <https://www.nature.com/articles/s41592-023-01766-6#Sec13>`_
    for details.

    """

    rng = np.random.default_rng(seed=random_state)
    original_labels = dataset.obsm[labels]
    adjacency_matrix = dataset.obsp["adjacency_matrix"]
    _, K = original_labels.shape

    original_enrichment = compute_neighborhood_enrichment(original_labels, adjacency_matrix)

    neighborhood_enrichments = np.zeros((n_trials, K, K), dtype=np.float64)
    for trial in range(n_trials):
        permuted_labels = rng.permutation(original_labels)
        neighborhood_enrichments[trial] = compute_neighborhood_enrichment(permuted_labels, adjacency_matrix)

    adjacency = (neighborhood_enrichments > original_enrichment[np.newaxis, :]).sum(axis=0)
    avoidance = (neighborhood_enrichments < original_enrichment[np.newaxis, :]).sum(axis=0)

    p_adjacency = (n_trials - adjacency + 1) / (n_trials + 1)
    p_avoidance = (n_trials - avoidance + 1) / (n_trials + 1)

    dataset.uns[f"adjacency_{pvalue_key}"] = p_adjacency
    dataset.uns[f"avoidance_{pvalue_key}"] = p_avoidance

    return dataset


def _adjacency_permutation_test(
    dataset: PopariDataset,
    labels: str = "X",
    n_trials: int = 100,
    random_state: int = 0,
    pvalue_key: str = "pvalue",
):
    r"""Compute p-values for neighborhood enrichment.

    See
    `In silico tissue generation and power analysis for spatial omics <https://www.nature.com/articles/s41592-023-01766-6#Sec13>`_
    for details.

    """

    rng = np.random.default_rng(seed=random_state)
    original_labels = dataset.obsm[labels]
    adjacency_matrix = dataset.obsp["adjacency_matrix"]
    _, K = original_labels.shape

    original_enrichment = compute_neighborhood_enrichment(original_labels, adjacency_matrix)

    neighborhood_enrichments = np.zeros((n_trials, K, K), dtype=np.float64)
    for trial in range(n_trials):
        permuted_labels = rng.permutation(original_labels)
        neighborhood_enrichments[trial] = compute_neighborhood_enrichment(permuted_labels, adjacency_matrix)

    adjacency = (neighborhood_enrichments > original_enrichment[np.newaxis, :]).sum(axis=0)
    avoidance = (neighborhood_enrichments < original_enrichment[np.newaxis, :]).sum(axis=0)

    p_adjacency = (n_trials - adjacency + 1) / (n_trials + 1)
    p_avoidance = (n_trials - avoidance + 1) / (n_trials + 1)

    dataset.uns[f"adjacency_{pvalue_key}"] = p_adjacency
    dataset.uns[f"avoidance_{pvalue_key}"] = p_avoidance

    return dataset


@enable_joint(annotations={"uns": ["ari"]})
@broadcast
def _compute_ari_score(dataset: PopariDataset, labels: str, predictions: str, ari_key: str = "ari"):
    r"""Compute adjusted Rand index (ARI) score  between a set of ground truth
    labels and an unsupervised clustering.

    Useful for assessing clustering validity. ARI score is computed per dataset.

    Args:
        dataset: dataset to process
        labels: the key in the ``.obs`` dataframe for the label data.
        predictions: the key in the ``.obs`` dataframe for the predictions data.
        ari_key: the key in the ``.uns`` dictionary where the ARI score will be stored.

    """

    ari = adjusted_rand_score(dataset.obs[labels], dataset.obs[predictions])
    dataset.uns[ari_key] = ari


@enable_joint(annotations={"uns": ["silhouette"]})
@broadcast
def _compute_silhouette_score(dataset: PopariDataset, labels: str, embeddings: str, silhouette_key: str = "silhouette"):
    r"""Compute silhouette score for a clustering based on Popari embeddings.

    Useful for assessing clustering validity. ARI score is computed per dataset.

    Args:
        dataset: dataset to process
        labels: the key in the ``.obs`` dataframe for the label data.
        predictions: the key in the ``.obs`` dataframe for the predictions data.
        ari_key: the key in the ``.uns`` dictionary where the ARI score will be stored.

    """

    silhouette = silhouette_score(dataset.obsm[embeddings], dataset.obs[labels])
    dataset.uns[silhouette_key] = silhouette


@broadcast
def _plot_all_embeddings(
    dataset: PopariDataset,
    embedding_key: str = "X",
    column_names: Optional[str] = None,
    **spatial_kwargs,
):
    r"""Plot all laerned metagenes in-situ across all replicates.

    Each replicate's metagenes are contained in a separate plot.

    Args:
        dataset: dataset to process
        embedding_key: the key in the ``.obsm`` dataframe for the cell/spot embeddings.
        column_names: a list of the suffixes for each latent feature. If ``None``, it is assumed
            that these suffixes are just the indices of the latent features.

    """

    _, K = dataset.obsm[f"{embedding_key}"].shape
    if column_names == None:
        # TODO: remove dependence on trained_model
        column_names = [f"{embedding_key}_{index}" for index in range(K)]

    edges_width = spatial_kwargs.pop("edges_width", 0.2)
    default_size = spatial_kwargs.pop("size", None)
    palette = spatial_kwargs.pop("palette", ListedColormap(sc.pl.palettes.godsnot_102))

    size = len(dataset) / 100
    if default_size is not None:
        size *= default_size

    axes = sq.pl.spatial_scatter(
        sq.pl.extract(dataset.copy(), embedding_key, prefix=f"{embedding_key}"),
        shape=None,
        color=column_names,
        edges_width=edges_width,
        connectivity_key="adjacency_matrix",
        size=size,
        wspace=0.2,
        ncols=2,
        **spatial_kwargs,
    )


@enable_joint(
    annotations={
        "uns": [
            "microprecision_train",
            "microprecision_validation",
            "macroprecision_train",
            "macroprecision_validation",
        ],
    },
)
@broadcast
def _evaluate_classification_task(dataset: PopariDataset, embeddings: str, labels: str):
    """"""

    le = LabelEncoder()
    encoded_labels = le.fit_transform(dataset.obs[labels].astype(str))
    dataset_embeddings = dataset.obsm[embeddings]

    X_train, X_valid, y_train, y_valid = train_test_split(
        dataset_embeddings,
        encoded_labels,
        train_size=0.25,
        random_state=42,
        stratify=encoded_labels,
    )
    model = KNeighborsClassifier(n_neighbors=10)
    model.fit(X_train, y_train)

    df = []
    for split, X, y in [("train", X_train, y_train), ("validation", X_valid, y_valid)]:
        y_soft = model.predict_proba(X)
        y_hat = np.argmax(y_soft, 1)
        dataset.uns[f"microprecision_{split}"] = precision_score(y, y_hat, average="micro")
        dataset.uns[f"macroprecision_{split}"] = precision_score(y, y_hat, average="macro")

    return dataset


@enable_joint(annotations={"obs": None, "uns": ["confusion_matrix"]})
@broadcast
def _compute_confusion_matrix(
    dataset: PopariDataset,
    labels: str,
    predictions: str,
    result_key: str = "confusion_matrix",
):
    r"""Compute confusion matrix for labels and predictions.

    Useful for visualizing clustering validity.

    Args:
        trained_model: the trained Popari model.
        labels: the key in the ``.obs`` dataframe for the label data.
        predictions: the key in the ``.obs`` dataframe for the predictions data.
        result_key: the key in the ``.uns`` dictionary where the reordered confusion matrix will be stored.

    """

    unique_labels = sorted(dataset.obs[labels].unique())
    unique_predictions = sorted(dataset.obs[predictions].unique())
    if len(unique_labels) != len(unique_predictions):
        raise ValueError("Number of unique labels and unique predictions must be equal.")

    encoded_labels = [unique_labels.index(label) for label in dataset.obs[labels].values]
    encoded_predictions = [unique_predictions.index(prediction) for prediction in dataset.obs[predictions].values]

    confusion_output = confusion_matrix(encoded_labels, encoded_predictions)

    permutation, index = get_optimal_permutation(confusion_output)
    dataset.obs[f"{labels}_inferred"] = [unique_labels[permutation[prediction]] for prediction in encoded_predictions]

    reordered_confusion = confusion_matrix(dataset.obs[labels], dataset.obs[f"{labels}_inferred"])[: len(unique_labels)]

    dataset.uns[result_key] = reordered_confusion


def get_optimal_permutation(confusion_output):
    """
    TODO: document
    maximum weight bipartite matching
    :param confusion_output:
    :return: confusion_output[perm, index], where index is sorted
    """

    num_label_classes, num_prediction_classes = confusion_output.shape

    bipartite_graph = nx.Graph()
    bipartite_graph.add_nodes_from([("label", i) for i in range(num_label_classes)], bipartite=0)
    bipartite_graph.add_nodes_from([("prediction", i) for i in range(num_prediction_classes)], bipartite=1)

    bipartite_graph.add_edges_from(
        [
            (("label", i), ("prediction", j), {"weight": confusion_output[i, j]})
            for i in range(num_label_classes)
            for j in range(num_prediction_classes)
        ],
    )

    assert nx.is_bipartite(bipartite_graph)
    matching = nx.max_weight_matching(bipartite_graph, maxcardinality=True)
    assert len({__ for _ in matching for __ in _}) == num_label_classes * 2

    matching = [sorted(match, key=lambda node_attributes: node_attributes[0]) for match in matching]

    matching = [tuple(index for (_, index) in match) for match in matching]
    matching = sorted(matching, key=lambda pair: pair[1])

    perm, index = tuple(map(np.array, zip(*matching)))

    return perm, index


@enable_joint(annotations={"uns": ["ground_truth_M_correlation"]})
@broadcast
def _compute_columnwise_autocorrelation(
    dataset: PopariDataset,
    uns: str = "ground_truth_M",
    result_key: str = "ground_truth_M_correlation",
):
    """"""

    matrix = dataset.uns[uns][f"{dataset.name}"].T

    num_columns, _ = matrix.shape
    correlation_coefficient_matrix = np.corrcoef(matrix, matrix)[:num_columns, :num_columns]
    dataset.uns[result_key] = correlation_coefficient_matrix


@broadcast_plottable
def _plot_confusion_matrix(
    dataset: PopariDataset,
    labels: str,
    ax=None,
    confusion_matrix_key: str = "confusion_matrix",
):
    ordered_labels = sorted(dataset.obs[labels].unique())
    sns.heatmap(
        dataset.uns[confusion_matrix_key],
        xticklabels=ordered_labels,
        yticklabels=ordered_labels,
        annot=True,
        ax=ax,
        fmt="3d",
    )
    plt.show()


@enable_joint(annotations={"uns": ["spatial_gene_correlation", "neighbor_interactions"]})
@broadcast
def _compute_spatial_gene_correlation(
    dataset: PopariDataset,
    spatial_key: str = "Sigma_x_inv",
    metagene_key: str = "M",
    spatial_gene_correlation_key: str = "spatial_gene_correlation",
    neighbor_interactions_key: str = "neighbor_interactions",
):
    """Computes spatial gene correlation according to learned metagenes."""

    spatial_affinity_matrix = dataset.uns[spatial_key][f"{dataset.name}"]
    metagenes = dataset.uns[metagene_key][f"{dataset.name}"]

    neighbor_interactions = metagenes @ spatial_affinity_matrix
    spatial_gene_correlation = neighbor_interactions @ metagenes.T

    dataset.uns[spatial_gene_correlation_key] = spatial_gene_correlation
    dataset.uns[neighbor_interactions_key] = neighbor_interactions


def _metagene_neighbor_interactions(dataset: PopariDataset, interaction_key: str = "metagene_neighbor_interactions"):
    """Compute pairwise interactions between every cell in terms of learned
    metagene embeddings.

    Can be used to visualize the empirical spatial correlations between metagenes.

    Args:
        dataset:

    """
    embeddings = dataset.obsm["X"]
    X = embeddings
    adjacency_matrix = dataset.obsp["adjacency_matrix"].toarray()

    adjacency_list = dataset.obsm["adjacency_list"]
    num_cells, num_metagenes = embeddings.shape

    Z = X / np.linalg.norm(X, axis=1, keepdims=True, ord=1)
    edges = np.array([(i, j) for i, e in enumerate(adjacency_list) for j in e])

    x = Z[edges[:, 0]]
    y = Z[edges[:, 1]]

    pair_interactions = np.zeros((num_cells, num_cells, num_metagenes, num_metagenes))
    cell_i, cell_j = adjacency_matrix.nonzero()
    for i in range(num_metagenes):
        for j in range(i, num_metagenes):
            pair_interactions[cell_i, cell_j, i, j] = 1 - x[:, i] * y[:, j]

    dataset.obsp[interaction_key] = pair_interactions


def _pretty_plot_spatial_affinities(
    datasets,
    metagene_key: str = "M",
    metagene_label_key: str = "metagene_labels",
    spatial_affinity_key: str = "Sigma_x_inv",
):
    """Rotate and plot spatial affinities matrix.

    Args:
        datasets:

    """

    transform_skew = Affine2D().skew_deg(15, 15)
    transform_rotate = Affine2D().rotate_deg(-45)
    transform = transform_skew + transform_rotate

    height = len(datasets) // 2 + len(datasets) % 2
    width = 2
    fig = plt.figure(dpi=1200, figsize=(width, height))

    first_dataset = datasets[0]
    _, K = first_dataset.uns[metagene_key][first_dataset.name].shape

    def setup_axes(fig, rect, metagene_ticks):
        """Setup axes for rotated heatmap plot."""

        grid_locator = FixedLocator([v for v, s in metagene_ticks])
        tick_formatter = DictFormatter(dict(metagene_ticks))

        grid_helper = floating_axes.GridHelperCurveLinear(
            transform,
            extremes=(-0.5, K - 0.5, -0.5, K - 0.5),  # TODO: adjust to fit number of metagenes K
            grid_locator1=grid_locator,
            tick_formatter1=tick_formatter,
            grid_locator2=grid_locator,
            tick_formatter2=tick_formatter,
        )
        ax = floating_axes.FloatingSubplot(fig, *rect, grid_helper=grid_helper)
        fig.add_subplot(ax)

        ax.axis["left"].toggle(ticklabels=False, label=False, ticks=False)

        ax.axis["right"].toggle(ticklabels=True, label=True, ticks=False)
        ax.axis["top"].toggle(ticks=False)

        ax.axis["bottom"].toggle(ticks=False)

        ax.axis["left"].line.set_linewidth(0)
        ax.axis["top"].line.set_linewidth(0)

        ax.axis["bottom"].line.set_linewidth(0.1)
        ax.axis["right"].line.set_linewidth(0.1)

        ax.axis["right"].major_ticklabels.set_axis_direction("right")
        ax.axis["bottom"].major_ticklabels.set_axis_direction("left")

        aux_ax = ax.get_aux_axes(transform)

        grid_helper.grid_finder.grid_locator1._nbins = 4
        grid_helper.grid_finder.grid_locator2._nbins = 4

        ax.axis[:].major_ticklabels.set_fontsize(2)

        return ax, aux_ax

    axes = []
    for index, dataset in enumerate(datasets):
        metagene_labels = dataset.uns[metagene_label_key]

        metagene_ticks = [(k, f"{metagene_labels[k]}" if k in metagene_labels else "N/A") for k in range(K)]

        rect = (height, width, index + 1)
        ax, aux_ax = setup_axes(fig, rect, metagene_ticks)
        ax.set_title(dataset.name, fontsize=2, y=0.5)
        axes.append(aux_ax)

    mask = np.ones((K, K), dtype=np.bool)
    mask[np.triu_indices_from(mask)] = 0

    _multireplicate_heatmap(
        datasets,
        uns=spatial_affinity_key,
        cmap="bwr",
        label_values=False,
        label_font_size=1.5,
        vmin=-10,
        vmax=10,
        mask=mask,
        axes=np.array(axes),
    )
    for ax in axes:
        im = ax.images
        cb = im[-1].colorbar
        cb.remove()

    fig.subplots_adjust(hspace=-0.5, wspace=-0.25)


def _plot_sparsity(dataset: ad.AnnData):
    """Plot overall sparsity of dataset.

    Args:
        dataset: dataset for which to compute and plot data sparsity.

    """

    raw_data = dataset.X

    if issparse(raw_data):
        raw_data = raw_data.toarray()

    raw_data = raw_data.flatten()
    raw_data_clipped = raw_data
    raw_data_clipped = raw_data_clipped[
        (raw_data_clipped > 1e-6) & (raw_data_clipped < np.percentile(raw_data_clipped, 99.9))
    ]

    fig, axes = plt.subplots(1, 2, sharey=True, tight_layout=True)
    axes[0].hist(raw_data, bins=20)
    axes[0].set_title("Count histogram")

    heights, *_ = axes[1].hist(raw_data_clipped, bins=20)
    axes[1].set_title("Clipped count histogram")
    axes[1].set_ylim([0, max(heights)])

    sparsity = (raw_data == 0).sum() / raw_data.size
    fig.suptitle(f"Overall sparsity: {sparsity}")

    return fig


def _plot_cell_type_to_metagene(
    dataset,
    cell_type_de_genes: dict,
    rank_mode: str = "metagene",
    plot_type: str = "box",
    normalize: bool = False,
    figsize: Optional[tuple] = None,
    cell_types: Optional[list] = None,
    metagene_index: Optional[Sequence] = None,
    alternative_hypothesis: str = "greater",
    metagene_key: str = "M",
    cmap: str = "rainbow",
    **subplot_kwargs,
):
    """Plot correspondence between cell types and metagenes.

    Args:
        dataset: output dataset from Popari
        cell_type_de_genes: mapping from cell types to DEG names.
        rank_mode: whether to compute ranks across metagenes or across genes.
        plot_type: whether to use box plot
        normalize: whether to normalize genes by zscore (across metagenes)
            before computing significance
        p_values: whether or not to calculate p-values for each cell type association
        cell_types: only plot listed cell types. Default: ``None`` (plot all cell types)
        metagene_index: only use listed metagenes types. Default: ``None`` (plot all cell types)
        metagene_key: key under which metagene values are stored. Default: ``M``

    """
    if cell_types is None:
        cell_types = cell_type_de_genes.keys()

    metagenes = dataset.uns[metagene_key][dataset.name]

    if normalize:
        metagenes = zscore(metagenes, axis=1)

    if metagene_index is None:
        num_genes, K = metagenes.shape
        metagene_index = np.arange(K)
    else:
        num_genes, _ = metagenes.shape
        K = len(metagene_index)

    sorted_indices = np.argsort(metagenes, axis=int(rank_mode != "metagene"))

    sorted_indices = sorted_indices[:, metagene_index]

    if figsize is None:
        figsize = (K, len(cell_types))

    fig, axes = plt.subplots(len(cell_types), 1, figsize=figsize, sharex=True, dpi=300, **subplot_kwargs, squeeze=False)

    ordered_genes = dataset.var_names
    #     axes[0].set_title("Gene Rank")

    means = {}
    for index, (cell_type, ax) in enumerate(zip(cell_types, axes.flat)):
        de_genes = cell_type_de_genes[cell_type]
        num_de_genes = len(de_genes)

        de_gene_indices = ordered_genes.get_indexer(de_genes)
        filtered_indices = [index for index in de_gene_indices if (index > -1)]
        num_actual_de_genes = len(filtered_indices)
        rank_distributions = []

        colors = colormaps[cmap].resampled(len(metagene_index))

        for sorted_index in sorted_indices.T:
            de_gene_ranks = sorted_index[filtered_indices]

            rank_distributions.append(de_gene_ranks)

        cell_type_means = [np.median(distribution) for distribution in rank_distributions]
        means[cell_type] = cell_type_means

        reduction = np.argmax if alternative_hypothesis == "greater" else np.argmin
        best_index = reduction(cell_type_means)

        best_distribution = rank_distributions[best_index]
        p_values = []
        for other_index in range(K):
            if other_index == best_index:
                p_values.append(0)
                continue
            other_distribution = rank_distributions[other_index]
            difference_distribution = best_distribution - other_distribution
            p_value = wilcoxon(difference_distribution, alternative=alternative_hypothesis).pvalue
            p_values.append(p_value)

        #         print(p_values)
        is_significant = (np.array(p_values) < (0.05 // len(p_values))).all()

        ax.text(1.05, 1, f"n={num_de_genes}[{num_actual_de_genes}]", transform=ax.transAxes, fontsize=figsize[0] + 2)

        if plot_type == "violin":
            plot = ax.violinplot(rank_distributions, showextrema=True, showmedians=True)

            plot["bodies"][0].set_zorder(4)
        elif plot_type == "box":
            medianprops = dict(linewidth=2, color="brown")
            plot = ax.boxplot(rank_distributions, widths=0.8, patch_artist=True, medianprops=medianprops)
            for median in plot["medians"]:
                median.set_zorder(1)
                median.set_label("Median rank")

            for index, patch in enumerate(plot["boxes"]):
                color = "white"
                if index != best_index:
                    p_value = p_values[index]
                    color = colors(index)
                    text_color = "red" if p_value <= 0.05 else "black"
                    ax.text(
                        index / K,
                        1,
                        f"p={p_value:.0E}",
                        rotation=30,
                        color=text_color,
                        transform=ax.transAxes,
                        fontsize=figsize[0] + 2,
                    )

                patch.set_facecolor(color)
                patch.set_zorder(0)

        ymax = num_genes if rank_mode == "metagene" else K
        ax.hlines(
            y=ymax // 2,
            xmin=0.5,
            xmax=K + 0.5,
            linewidth=1.5,
            linestyle="--",
            color="k",
            zorder=2,
            label="Null hypothesis rank",
        )

        # Setting axis tick lines
        tick_interval = 1000
        top_tick = num_genes // tick_interval
        y_ticks = [1] + [(i + 1) * tick_interval for i in range(top_tick)]
        y_ticklabels = [y_tick if y_tick in (1, top_tick * tick_interval) else "" for y_tick in y_ticks]
        ax.set_yticks(y_ticks)
        ax.set_yticklabels(y_ticklabels)

        ax.tick_params(axis="x", bottom=False)
        ax.set_ylabel(cell_type)

        ax.yaxis.tick_right()
        for side, spine in ax.spines.items():
            if side != "right":
                spine.set_edgecolor("none")
            else:
                spine.set_bounds((1, num_genes))

    ax.set_xticks(np.arange(len(metagene_index)) + 1)
    ax.set_xticklabels([f"m{m}" for m in metagene_index])

    plt.subplots_adjust(left=None, bottom=None, right=None, top=None, wspace=None, hspace=0.1)

    lines_labels = [ax.get_legend_handles_labels() for ax in fig.axes]
    lines, labels = (sum(lol, []) for lol in zip(*lines_labels))

    # Create legend for null hypothesis and medians of boxplots
    filtered_lines = []
    filtered_labels = []
    for label in np.unique(labels):
        first_index = labels.index(label)
        filtered_labels.append(label)
        filtered_lines.append(lines[first_index])

    fig.legend(filtered_lines, filtered_labels, loc="upper left", bbox_to_anchor=(1, 1))

    fig.supxlabel("Metagene")
    fig.supylabel("Cell Type")

    return fig, means


def _plot_cell_type_to_metagene_difference(
    dataset,
    cell_type_de_genes: dict,
    first_metagene: int,
    second_metagene: int,
    rank_mode: str = "metagene",
    plot_type: str = "box",
    normalize: bool = False,
    figsize: Optional[tuple] = None,
    cell_types: Optional[list] = None,
    alternative_hypothesis: str = "greater",
    metagene_key: str = "M",
    cmap: str = "rainbow",
    **subplot_kwargs,
):
    """Plot correspondence between cell types and metagenes.

    Args:
        dataset: output dataset from Popari
        cell_type_de_genes: mapping from cell types to DEG names.
        rank_mode: whether to compute ranks across metagenes or across genes.
        plot_type: whether to use box plot
        normalize: whether to normalize genes by zscore (across metagenes)
            before computing significance
        p_values: whether or not to calculate p-values for each cell type association
        cell_types: only plot listed cell types. Default: ``None`` (plot all cell types)
        metagene_index: only use listed metagenes types. Default: ``None`` (plot all cell types)
        metagene_key: key under which metagene values are stored. Default: ``M``

    """
    if cell_types is None:
        cell_types = cell_type_de_genes.keys()

    metagenes = dataset.uns[metagene_key][dataset.name]

    if normalize:
        metagenes = zscore(metagenes, axis=1)

    num_genes, K = metagenes.shape

    sorted_indices = np.argsort(metagenes, axis=int(rank_mode != "metagene"))

    if figsize is None:
        figsize = (len(cell_types), 4)

    fig, ax = plt.subplots(figsize=figsize, sharex=True, dpi=300, **subplot_kwargs)

    ordered_genes = dataset.var_names
    #     axes[0].set_title("Gene Rank")

    means = {}
    difference_distributions = []
    p_values = []
    for index, cell_type in enumerate(cell_types):
        de_genes = cell_type_de_genes[cell_type]
        num_de_genes = len(de_genes)

        de_gene_indices = ordered_genes.get_indexer(de_genes)
        filtered_indices = [index for index in de_gene_indices if (index > -1)]
        num_actual_de_genes = len(filtered_indices)

        colors = colormaps[cmap].resampled(2)

        first_metagene_de_gene_ranks = sorted_indices[:, first_metagene][filtered_indices]
        second_metagene_de_gene_ranks = sorted_indices[:, second_metagene][filtered_indices]

        difference_distribution = first_metagene_de_gene_ranks - second_metagene_de_gene_ranks
        difference_distributions.append(difference_distribution)

        p_value = wilcoxon(difference_distribution, alternative=alternative_hypothesis).pvalue
        p_values.append(p_value)

    cell_type_means = [np.median(distribution) for distribution in difference_distributions]
    means[cell_type] = cell_type_means

    #         print(p_values)
    is_significant = (np.array(p_values) < (0.05 // len(p_values))).all()

    # ax.text(1.05, 1, f"n={num_de_genes}[{num_actual_de_genes}]", transform=ax.transAxes, fontsize=figsize[0] + 2)

    if plot_type == "violin":
        plot = ax.violinplot(rank_distributions, showextrema=True, showmedians=True)

        plot["bodies"][0].set_zorder(4)
    elif plot_type == "box":
        medianprops = dict(linewidth=2, color="brown")
        plot = ax.boxplot(difference_distributions, widths=0.8, patch_artist=True, medianprops=medianprops)
        for median in plot["medians"]:
            median.set_zorder(1)
            median.set_label("Median difference")

        for index, patch in enumerate(plot["boxes"]):
            p_value = p_values[index]
            color = colors(index)
            text_color = "red" if p_value <= 0.05 else "black"
            ax.text(
                index / len(cell_types),
                1,
                f"p={p_value:.0E}",
                rotation=30,
                color=text_color,
                transform=ax.transAxes,
                fontsize=figsize[0] + 2,
            )

            patch.set_facecolor(color)
            patch.set_zorder(0)

    ax.hlines(
        y=0,
        xmin=0.5,
        xmax=len(cell_types) + 1,
        linewidth=1.5,
        linestyle="--",
        color="k",
        zorder=2,
        label="Null hypothesis rank",
    )

    # Setting axis tick lines
    tick_interval = 1000
    top_tick = num_genes // tick_interval
    y_ticks = [i * tick_interval for i in range(-top_tick, top_tick + 1)]
    y_ticklabels = [
        y_tick if y_tick in (-top_tick * tick_interval, 0, top_tick * tick_interval) else "" for y_tick in y_ticks
    ]
    ax.set_yticks(y_ticks)
    ax.set_yticklabels(y_ticklabels)

    ax.tick_params(axis="x", bottom=False)
    ax.set_ylabel(f"Difference distribution between m{first_metagene} and m{second_metagene}")

    ax.yaxis.tick_right()
    for side, spine in ax.spines.items():
        if side != "right":
            spine.set_edgecolor("none")
        else:
            pass
            spine.set_bounds((-num_genes, num_genes))

    ax.set_xticks(np.arange(len(cell_types)) + 1)
    ax.set_xticklabels(cell_types, rotation=-30, ha="left")

    plt.subplots_adjust(left=None, bottom=None, right=None, top=None, wspace=None, hspace=0.1)

    lines_labels = [ax.get_legend_handles_labels() for ax in fig.axes]
    lines, labels = (sum(lol, []) for lol in zip(*lines_labels))

    # Create legend for null hypothesis and medians of boxplots
    filtered_lines = []
    filtered_labels = []
    for label in np.unique(labels):
        first_index = labels.index(label)
        filtered_labels.append(label)
        filtered_lines.append(lines[first_index])

    fig.legend(filtered_lines, filtered_labels, loc="upper left", bbox_to_anchor=(1, 1))

    ax.set_xlabel("Cell Type")

    return fig, means


def _compile_de_genes(
    dataset,
    de_category: str = "cell_type",
    filtered_deg_key: str = "t-test filtered",
    p_value_threshold: float = 1e-5,
    max_genes: int = 500,
    format="disk",
):
    """Compile DE genes.

    After running ``sc.tl.rank_genes_groups`` and ``sc.tl.filter_rank_genes_groups``, call this
    function in order to summarize and collect the DE genes identified before.

    Args:
        dataset:

    """
    all_de_genes = set()
    ranked_genes = dataset.uns[filtered_deg_key]["names"]

    # Record arrays are nasty, and will be removed in Scanpy 2.0, probably
    for cell_type in ranked_genes.dtype.names:
        index = ranked_genes[cell_type] == "nan"
        ranked_genes[cell_type][index] = np.nan

    adjusted_pvals = dataset.uns[filtered_deg_key]["pvals_adj"]
    cell_type_de_genes = {}
    for cell_type in dataset.obs[de_category].unique():
        cell_type_genes = ranked_genes[cell_type]
        cell_type_pvals = adjusted_pvals[cell_type]

        pval_filter = cell_type_pvals < p_value_threshold

        filtered_genes = cell_type_genes[(pval_filter) & (~pd.isna(cell_type_genes))][:max_genes]
        cell_type_de_genes[cell_type] = filtered_genes
        print(cell_type, len(filtered_genes))
        all_de_genes.update(filtered_genes)

    if format == "disk":
        for cell_type in ranked_genes.dtype.names:
            index = pd.isna(ranked_genes[cell_type])
            ranked_genes[cell_type][index] = "nan"
        # ranked_genes[cell_type][index] = ranked_genes[cell_type][index].astype(str)

    print(len(all_de_genes))

    return cell_type_de_genes, all_de_genes


def _call_de_genes(
    dataset,
    label_key: str,
    output_key: str,
    filter: bool = True,
    min_fold_change: float = 1,
    max_genes: int = 500,
):
    """Call differentially expressed genes using a t-test, and compile.

    Args:

    """
    filtered_deg_key = output_key

    sc.tl.rank_genes_groups(dataset, label_key, method="t-test", key_added=output_key, min_fold_change=min_fold_change)
    if filter:
        filtered_deg_key = f"{filtered_deg_key}_filtered"
        sc.tl.filter_rank_genes_groups(
            dataset,
            key=output_key,
            key_added=f"{output_key}_filtered",
            min_fold_change=min_fold_change,
        )

    cell_type_de_genes, all_de_genes = _compile_de_genes(
        dataset,
        label_key,
        max_genes=max_genes,
        filtered_deg_key=filtered_deg_key,
    )

    return cell_type_de_genes, all_de_genes


@enable_joint
@broadcast_plottable
def _plot_embeddings_to_label(
    dataset,
    ax=None,
    names: Optional[Sequence[str]] = None,
    embedding_key: str = "normalized_X",
    label_key: str = "leiden",
    **dotplot_kwargs,
):
    """Plot enrichment of embedding dimensions for a label category.

    Args:
        names: Names of embedding dimensions
        swap

    """

    X = dataset.obsm[embedding_key]
    num_spots, K = X.shape

    mock_dataset = ad.AnnData(
        X=dataset.obsm[embedding_key],
    )

    if not names:
        mock_dataset.var_names = [f"m{index}" for index in range(K)]
    else:
        mock_dataset.var_names = names

    swap_axes = dotplot_kwargs.pop("swap_axes", True)
    standard_scale = dotplot_kwargs.pop("standard_scale", "var")

    mock_dataset.obs = dataset.obs.copy()

    aggregated = sc.get.aggregate(
        mock_dataset,
        by=label_key,
        func=["sum", "count_nonzero"],
    )

    mean_in_expressed = aggregated.to_df(layer="sum") / aggregated.to_df(layer="count_nonzero")

    vmin = None
    vmax = None
    if standard_scale is None:
        vmax = np.max(np.abs(mean_in_expressed))
        vmin = -vmax

    dotplot = sc.pl.dotplot(
        mock_dataset,
        mock_dataset.var_names,
        groupby=label_key,
        dendrogram=False,
        ax=ax,
        standard_scale=standard_scale,
        swap_axes=swap_axes,
        return_fig=True,
        vmin=vmin,
        vmax=vmax,
        **dotplot_kwargs,
    )

    dotplot.add_totals()
    dotplot.show()

    return dotplot, mean_in_expressed


@enable_joint(annotations={"obsm": ["marker_expression"]})
@broadcast
def _score_marker_expression(dataset, de_genes: dict[str, Sequence[str]], output_key="marker_expression"):
    """Given a mapping from cell types to marker genes, compute enrichment."""
    data = dataset.X if not issparse(dataset.X) else dataset.X.todense()
    zscored_expression = zscore(data)

    marker_gene_expression = np.zeros((len(dataset), len(de_genes)))
    for index, (subtype, gene_list) in enumerate(de_genes.items()):
        gene_list_index = dataset.var_names.isin(gene_list)
        marker_gene_expression[:, index] = zscored_expression[:, gene_list_index].sum(axis=1)

    dataset.obsm[output_key] = marker_gene_expression


@enable_joint(annotations={"obsm": ["marker_expression"]})
def _plot_clusters_to_categories(
    datasets,
    category_de_genes: dict[str, Sequence[str]],
    output_key="marker_expression",
    **dotplot_kwargs,
):
    """Plot clusters to category correspondence, based on category marker gene
    expression.

    Args:
        category_de_genes: mapping from category to marker genes for that category

    """
    _score_marker_expression.__wrapped__(datasets, category_de_genes, output_key=output_key)
    (cell_type_to_cluster_correspondence_fig, (_, mean_in_expressed)) = _plot_embeddings_to_label.__wrapped__(
        datasets,
        names=list(category_de_genes.keys()),
        standard_scale=None,
        cmap="bwr",
        embedding_key=output_key,
        title="Cell-type-to-cluster Correspondence",
        **dotplot_kwargs,
    )

    return cell_type_to_cluster_correspondence_fig


def _cluster_domains(
    datasets,
    target_domains: Optional[int] = None,
    skip_thresholding: bool = True,
    batch_correct: bool = False,
):
    """Discover domains from Popari embeddings."""

    normalized_key = "normalized_X"
    if not skip_thresholding:
        normalized_key = "normalized_thresholded_expression"
        for dataset in datasets:
            normalize_expression_by_threshold(dataset, thresholded_key="normalized_X")

    processed_key = normalized_key
    if batch_correct:
        merged_dataset = concatenate(datasets)
        sce.pp.scanorama_integrate(merged_dataset, "batch", basis=normalized_key, verbose=1)

        processed_key = "X_scanorama"
        for dataset in datasets:
            dataset.obsm[processed_key] = (
                merged_dataset[merged_dataset.obs["batch"] == dataset.name].obsm[processed_key].copy()
            )

    for dataset in datasets:
        smooth_metagene_expression(dataset, processed_key=processed_key)

    _cluster(
        datasets,
        verbose=True,
        use_rep="smoothed_expression",
        target_clusters=target_domains,
        n_neighbors=40,
        joint=True,
    )

    for dataset in datasets:
        smooth_labels(dataset, output_key="smoothed_domain")


def _metagene_gsea(
    dataset: PopariDataset,
    metagene_index: int,
    metagene_key: str = "M",
    sensitivity: float = 0.5,
    **gsea_kwargs,
):
    """Compute GSEA for the top genes from a particular metagene."""
    metagenes = dataset.uns[metagene_key][dataset.name]
    metagene = metagenes[:, metagene_index]

    gene_names = dataset.var_names

    signature = list(get_metagene_signature(metagene, gene_names, show_plot=False, sensitivity=sensitivity))

    mode = gsea_kwargs.pop("mode", "barplot")
    try:
        enrichment_result, ax = run_gsea(
            signature,
            f"Metagene {metagene_index}",
            gene_names,
            mode=mode,
        )

        fig = ax.figure
    except ValueError:
        raise ValueError("Didn't find any enriched GO terms.")

    return fig


def _plot_metagene_signature_enrichment(
    datasets,
    metagene_index,
    categories=None,
    category_key="domain",
    sensitivity=1,
    **subplot_kwargs,
):
    """Plot enrichment of metagene signature across categories."""

    first_dataset = datasets[0]
    M = list(first_dataset.uns["M"].values())[0]

    gene_names = first_dataset.var_names

    merged_dataset = concatenate(datasets)
    if categories is None:
        categories = merged_dataset.obs[category_key].unique()

    metagene_signature = get_metagene_signature(
        M[:, metagene_index],
        gene_names,
        sensitivity=sensitivity,
        type="upregulated",
        show_plot=False,
    )

    indices = gene_names.get_indexer(metagene_signature)

    num_columns = len(datasets) * len(categories)
    raw_genes = np.zeros((len(metagene_signature), num_columns))

    for category_index, category in enumerate(categories):
        for dataset_index, dataset in enumerate(datasets):

            first_subtype = dataset[dataset.obs[category_key] == category]
            raw_genes[:, dataset_index * len(categories) + category_index] = first_subtype.X[:, indices].mean(axis=0)

    vmax = np.abs(raw_genes).max()
    vmin = 0

    dpi = subplot_kwargs.pop("dpi", 200)
    figsize = subplot_kwargs.pop("figsize", (4, len(metagene_signature) / 5))
    fig, ax = plt.subplots(dpi=dpi, figsize=figsize, **subplot_kwargs)
    im = ax.imshow(raw_genes, cmap="Reds", vmax=vmax, vmin=vmin)

    for index in range(0, len(datasets)):
        ax.axvline((index) * len(categories) - 0.5, color="k")

    ax.grid(False)

    plt.colorbar(im, ax=ax)
    ax.set_yticks(np.arange(len(metagene_signature)))
    ax.set_yticklabels(metagene_signature)
    ax.set_xticks(np.arange(num_columns))
    ax.set_xticklabels([])
    ax.set_aspect("equal")
    colors = plt.get_cmap("tab20b").colors  # Colors for each tick

    for i, tick in enumerate(ax.get_xticks()):
        rect = patches.Rectangle(
            ((tick + 0.1) / num_columns, 0),
            0.04,
            -0.5 / len(metagene_signature),
            transform=ax.transAxes,
            color=colors[i % len(categories)],
        )
        fig.add_artist(rect)

    return fig, metagene_signature



from __future__ import annotations

from collections import defaultdict
from typing import Optional, Sequence

import numpy as np
import torch
from scipy.sparse import csr_array
from tqdm.auto import trange

from popari._binning_utils import GridDownsampler, PartitionDownsampler
from popari._embedding_optimizer import EmbeddingOptimizer
from popari._parameter_optimizer import ParameterOptimizer
from popari._popari_dataset import PopariDataset
from popari.initialization import initialize_dummy, initialize_kmeans, initialize_leiden, initialize_svd
from popari.sample_for_integral import integrate_of_exponential_over_simplex
from popari.util import convert_numpy_to_pytorch_sparse_coo, get_datetime


class HierarchicalView:
    """View of SRT multisample dataset at a set resolution.

    Includes the scaled (i.e. binned data) as well as the learnable Popari
    parameters and their corresponding optimizers.

    """

    def __init__(
        self,
        datasets: Sequence[PopariDataset],
        betas: list,
        prior_x_modes: list,
        method: str,
        random_state: int,
        K: int,
        context: dict,
        initial_context: dict,
        use_inplace_ops: bool,
        pretrained: bool,
        verbose: str,
        metagene_groups: dict,
        spatial_affinity_groups: dict,
        parameter_optimizer_hyperparameters: dict,
        embedding_optimizer_hyperparameters: dict,
        binned_Ys: list = None,
        superresolution_lr: float = 1e-3,
        level: int = 0,
        hierarchical_levels: int | None = 1,
    ):

        self.datasets = datasets
        self.replicate_names = [dataset.name for dataset in datasets]
        self.K = K
        self.level = level
        self.hierarchical_levels = hierarchical_levels
        self.level_suffix = "" if self.level == 0 else f"_level_{self.level}"
        self.context = context
        self.initial_context = initial_context
        self.use_inplace_ops = use_inplace_ops
        self.verbose = verbose
        self.random_state = random_state
        self.pretrained = pretrained

        self.num_replicates = len(self.datasets)

        def fill_groups(groups, are_exclusive=False):
            if not groups:
                groups = {f"_default": self.replicate_names}

            included_replicate_names = sum(groups.values(), [])
            difference = set(self.replicate_names) - set(included_replicate_names)
            if difference:
                groups[f"_default"] = list(difference)

            # Make group names unique to hierarchical level
            groups = {
                key if key.endswith(self.level_suffix) else f"{key}{self.level_suffix}": value
                for key, value in groups.items()
            }

            tags = {replicate_name: [] for replicate_name in self.replicate_names}
            for group, group_replicates in groups.items():
                for replicate in group_replicates:
                    if are_exclusive and len(tags[replicate]) > 0:
                        ValueError("If in shared mode, each replicate can only appear in one group.")
                    tags[replicate].append(group)

            return groups, tags

        def add_level_suffix(groups):
            suffixed_groups = {}
            for group, group_replicates in groups.items():
                suffixed_groups[group] = [
                    f"{group_replicate}{self.level_suffix}" for group_replicate in group_replicates
                ]

            return suffixed_groups

        if metagene_groups == "disjoint":
            metagene_groups = {replicate_name: [replicate_name] for replicate_name in self.replicate_names}

        if spatial_affinity_groups == "disjoint":
            spatial_affinity_groups = {replicate_name: [replicate_name] for replicate_name in self.replicate_names}

        if metagene_groups is not None:
            metagene_groups = add_level_suffix(metagene_groups)
        if spatial_affinity_groups is not None:
            spatial_affinity_groups = add_level_suffix(spatial_affinity_groups)

        self.metagene_groups, self.metagene_tags = fill_groups(
            metagene_groups,
            are_exclusive=(parameter_optimizer_hyperparameters["metagene_mode"] == "shared"),
        )
        self.spatial_affinity_groups, self.spatial_affinity_tags = fill_groups(
            spatial_affinity_groups,
            are_exclusive=(parameter_optimizer_hyperparameters["spatial_affinity_mode"] == "shared lookup"),
        )

        parameter_optimizer_hyperparameters["metagene_groups"] = self.metagene_groups
        parameter_optimizer_hyperparameters["metagene_tags"] = self.metagene_tags
        parameter_optimizer_hyperparameters["spatial_affinity_groups"] = self.spatial_affinity_groups
        parameter_optimizer_hyperparameters["spatial_affinity_tags"] = self.spatial_affinity_tags

        if binned_Ys is None:
            self.Ys = []
            for dataset in self.datasets:
                num_cells, _ = dataset.shape
                Y = convert_numpy_to_pytorch_sparse_coo(dataset.X, self.context)
                Y *= (self.K * 1) / (Y.sum() / num_cells)
                self.Ys.append(Y)
        else:
            self.Ys = binned_Ys

        if betas is None:
            self.betas = np.full(self.num_replicates, 1 / self.num_replicates)
        else:
            self.betas = np.array(betas, copy=False) / sum(betas)

        if prior_x_modes is None:
            prior_x_modes = [None] * self.num_replicates

        self.prior_x_modes = prior_x_modes

        self.parameter_optimizer = ParameterOptimizer(
            self.K,
            self.Ys,
            self.datasets,
            self.betas,
            prior_x_modes,
            initial_context=self.initial_context,
            context=self.context,
            use_inplace_ops=self.use_inplace_ops,
            verbose=self.verbose,
            **parameter_optimizer_hyperparameters,
        )
        self.superresolution_lr = superresolution_lr

        if self.verbose:
            print(f"{get_datetime()} Initializing EmbeddingOptimizer")
        self.embedding_optimizer = EmbeddingOptimizer(
            self.K,
            self.Ys,
            self.datasets,
            initial_context=self.initial_context,
            context=self.context,
            use_inplace_ops=self.use_inplace_ops,
            verbose=self.verbose,
            **embedding_optimizer_hyperparameters,
        )
        self.parameter_optimizer.link(self.embedding_optimizer)
        self.embedding_optimizer.link(self.parameter_optimizer)

        if self.pretrained:
            first_dataset = self.datasets[0]
            # if self.metagene_mode == "differential":
            #     self.parameter_optimizer.metagene_state.M_bar = {group_name: torch.from_numpy(first_dataset.uns["M_bar"][group_name]).to(**self.initial_context) for group_name in self.metagene_groups}
            # if self.spatial_affinity_mode == "differential lookup":
            #     self.parameter_optimizer.spatial_affinity_state.spatial_affinity_bar = {group_name: first_dataset.uns["M_bar"][group_name].to(**self.initial_context) for group_name in self.spatial_affinity_groups}
            spatial_affinity_copy = torch.zeros((len(self.datasets), self.K, self.K), **self.context)
            for dataset_index, dataset in enumerate(self.datasets):
                self.parameter_optimizer.metagene_state[dataset.name][:] = torch.from_numpy(
                    dataset.uns["M"][dataset.name],
                ).to(**self.initial_context)
                self.embedding_optimizer.embedding_state[dataset.name][:] = torch.from_numpy(dataset.obsm["X"]).to(
                    **self.initial_context,
                )
                self.embedding_optimizer.adjacency_matrices[dataset.name] = convert_numpy_to_pytorch_sparse_coo(
                    dataset.obsp["adjacency_matrix"],
                    self.initial_context,
                )
                self.parameter_optimizer.adjacency_matrices[dataset.name] = convert_numpy_to_pytorch_sparse_coo(
                    dataset.obsp["adjacency_matrix"],
                    self.initial_context,
                )

                self.parameter_optimizer.spatial_affinity_state[dataset.name] = torch.from_numpy(
                    dataset.uns["Sigma_x_inv"][dataset.name],
                ).to(**self.initial_context)
                spatial_affinity_copy[dataset_index] = self.parameter_optimizer.spatial_affinity_state[dataset.name]

            self.parameter_optimizer.update_sigma_yx()
            self.parameter_optimizer.spatial_affinity_state.initialize_optimizers(spatial_affinity_copy)
        else:
            if self.level < self.hierarchical_levels - 1:
                method = "dummy"

            if self.verbose:
                print(f"{get_datetime()} Initializing metagenes and hidden states using {method} method")

            if method == "dummy":
                self.M, self.Xs = initialize_dummy(self.datasets, self.K, self.initial_context)
            elif method == "kmeans":
                self.M, self.Xs = initialize_kmeans(
                    self.datasets,
                    self.K,
                    self.initial_context,
                    kwargs_kmeans=dict(random_state=self.random_state),
                )
            elif method == "svd":
                self.M, self.Xs = initialize_svd(
                    self.datasets,
                    self.K,
                    self.initial_context,
                    M_nonneg=(self.parameter_optimizer.M_constraint == "simplex"),
                    X_nonneg=True,
                )
            elif method == "leiden":
                kwargs_leiden = {
                    "random_state": self.random_state,
                }
                self.M, self.Xs = initialize_leiden(
                    self.datasets,
                    self.K,
                    self.initial_context,
                    kwargs_leiden=kwargs_leiden,
                    verbose=self.verbose,
                )
            else:
                raise NotImplementedError

            for dataset_index, dataset in enumerate(self.datasets):
                self.parameter_optimizer.metagene_state[dataset.name][:] = self.M
                self.embedding_optimizer.embedding_state[dataset.name][:] = self.Xs[dataset_index]

            self.parameter_optimizer.scale_metagenes()

            # # Ensure initial embeddings do not have too large magnitudes
            # for dataset_index, dataset in enumerate(self.datasets):
            #     initial_X = self.embedding_optimizer.embedding_state[dataset.name]
            #     cell_normalized_X = initial_X / torch.linalg.norm(initial_X, dim=0, keepdim=True)
            #     self.embedding_optimizer.embedding_state[dataset.name][:] = cell_normalized_X

            self.Sigma_x_inv_bar = None

            self.parameter_optimizer.update_sigma_yx()

            # # Update metagenes to ensure that they lie on simplex after normalizign embeddings
            # self.parameter_optimizer.update_metagenes()

            initial_embeddings = [self.embedding_optimizer.embedding_state[dataset.name] for dataset in self.datasets]

            # Initializing spatial affinities
            if self.verbose:
                print(f"{get_datetime()} Initializing Sigma_x_inv with empirical correlations")
            self.parameter_optimizer.spatial_affinity_state.initialize(initial_embeddings)

            for dataset_index, dataset in enumerate(self.datasets):
                metagene_state = self.parameter_optimizer.metagene_state[dataset.name].cpu().detach().numpy()
                dataset.uns["M"] = {dataset.name: metagene_state}

                X = self.embedding_optimizer.embedding_state[dataset.name].cpu().detach().numpy()
                dataset.obsm["X"] = X

                Sigma_x_inv = self.parameter_optimizer.spatial_affinity_state[dataset.name].cpu().detach().numpy()
                dataset.uns["Sigma_x_inv"] = {dataset.name: Sigma_x_inv}

                dataset.uns["popari_hyperparameters"] = {
                    "prior_x": self.parameter_optimizer.prior_xs[dataset_index][0].cpu().detach().numpy(),
                    "K": self.K,
                    "use_inplace_ops": self.use_inplace_ops,
                    "random_state": self.random_state,
                    "verbose": self.verbose,
                    **parameter_optimizer_hyperparameters,
                    **embedding_optimizer_hyperparameters,
                }

                # if self.metagene_mode == "differential":
                #     dataset.uns["popari_hyperparameters"]["lambda_M"] = self.parameter_optimizer.lambda_M
                # if self.spatial_affinity_mode != "shared":
                #     dataset.uns["popari_hyperparameters"]["lambda_Sigma_bar"] = self.parameter_optimizer.lambda_Sigma_bar

            if self.parameter_optimizer.metagene_mode == "differential":
                M_bar = {
                    group_name: self.parameter_optimizer.metagene_state.M_bar[group_name].cpu().detach().numpy()
                    for group_name in self.parameter_optimizer.metagene_groups
                }
                for dataset in self.datasets:
                    dataset.uns["M_bar"] = M_bar
                    # dataset.uns["lambda_Sigma_bar"] = self.parameter_optimizer.lambda_Sigma_bar

            if self.parameter_optimizer.spatial_affinity_mode == "differential lookup":
                spatial_affinity_bar = {
                    group_name: self.parameter_optimizer.spatial_affinity_state.spatial_affinity_bar[group_name]
                    .cpu()
                    .detach()
                    .numpy()
                    for group_name in self.parameter_optimizer.spatial_affinity_groups
                }
                for dataset in self.datasets:
                    dataset.uns["spatial_affinity_bar"] = spatial_affinity_bar
                    # dataset.uns["lambda_M"] = self.parameter_optimizer.lambda_M

        self.superresolution_optimizers = {}
        for dataset in self.datasets:
            if "losses" not in dataset.uns:
                dataset.uns["losses"] = defaultdict(list)
            else:
                dataset.uns["losses"] = defaultdict(list, dataset.uns["losses"])
                for key in dataset.uns["losses"]:
                    dataset.uns["losses"][key] = list(dataset.uns["losses"][key])

        if self.parameter_optimizer.metagene_mode == "differential":
            self.parameter_optimizer.metagene_state.reaverage()
        if self.parameter_optimizer.spatial_affinity_mode == "differential lookup":
            self.parameter_optimizer.spatial_affinity_state.reaverage()

    def link(self, low_res_view: HierarchicalView):
        """Link a view to the resolution right below it in the hierarchy."""
        self.low_res_view = low_res_view

    def _propagate_parameters(self):
        """Use parameters from low-resolution to initialize higher-
        resolution."""
        low_res_metagenes = self.low_res_view.parameter_optimizer.metagene_state.metagenes
        self.parameter_optimizer.metagene_state.metagenes[:] = low_res_metagenes

        self.synchronize_datasets()

    def _superresolve_embeddings(
        self,
        n_epochs=10000,
        tol=1e-4,
        update_alg="gd",
        use_manual_gradients=True,
        verbose=None,
    ):
        """Superresolve embeddings using embeddings for lower resolution
        spots."""

        final_losses = np.zeros(len(self.datasets))
        for dataset_index, (dataset, low_res_name) in enumerate(zip(self.datasets, self.low_res_view.replicate_names)):
            low_res_dataset = self.low_res_view.datasets[dataset_index]
            sigma_yx = self.parameter_optimizer.sigma_yxs[dataset_index]
            Y = self.Ys[dataset_index].to(self.context["device"])

            if Y.sum() == 0:
                raise ValueError(
                    "It seems like you are trying to superresolve a hierarchical level with all zero expression "
                    "values. This probably means the model was saved incorrectly; try using `model.save_results` "
                    "with `as_trainable=True` next time.",
                )

            X = self.embedding_optimizer.embedding_state[dataset.name].to(self.context["device"])
            X_B = low_res_dataset.obsm["X"]
            B = low_res_dataset.obsm[f"bin_assignments_{low_res_dataset.name}"]

            M = self.parameter_optimizer.metagene_state[dataset.name].to(self.context["device"])
            prior_x_mode = self.parameter_optimizer.prior_x_modes[dataset_index]
            prior_x = self.parameter_optimizer.prior_xs[dataset_index]

            # Precomputing quantities
            MTM = M.T @ M / (sigma_yx**2)
            BTB = convert_numpy_to_pytorch_sparse_coo((B.T @ B).tocoo(), context=self.context)
            YM = Y @ M / (sigma_yx**2)
            BTX_B = torch.from_numpy(B.T @ X_B).to(self.context["device"]).to(self.context["dtype"])

            linear_term_gradient = YM + BTX_B
            if prior_x_mode == "exponential shared fixed":
                linear_term_gradient = linear_term_gradient - prior_x[0][None]

            Ynorm = torch.square(Y).sum() / (sigma_yx**2)
            X_Bnorm = np.linalg.norm(X_B, ord="fro").item() ** 2
            loss_prev, loss = np.inf, np.nan

            X = X.clone().detach().requires_grad_(True)

            if verbose is None:
                verbose = self.verbose

            superresolution_optimizer = torch.optim.Adam(
                [X],
                lr=self.superresolution_lr,
                betas=(0.5, 0.9),
            )
            # superresolution_scheduler = torch.optim.lr_scheduler.StepLR(superresolution_optimizer, step_size=decay_period, gamma=0.5)

            # self.superresolution_optimizers[dataset.name] = superresolution_optimizer
            # self.superresolution_schedulers[dataset.name] = superresolution_scheduler

            def gradient_update(X, iteration=None):
                """TODO:UNTESTED."""

                # lambda_B = 1e-1
                # self.superresolution_optimizers[dataset.name].zero_grad()
                superresolution_optimizer.zero_grad()
                # X.grad = None
                quadratic_term_gradient = X @ MTM + BTB @ X

                loss = (
                    (quadratic_term_gradient * X).sum() / 2 - (linear_term_gradient * X).sum() + Ynorm / 2 + X_Bnorm / 2
                )
                if use_manual_gradients:
                    gradient = quadratic_term_gradient - linear_term_gradient
                    X.grad = gradient
                else:
                    # loss = (torch.linalg.norm(Y.T - M @ X.T, ord='fro') ** 2) / 2  + ((torch.linalg.norm(X_B - B @ X, ord='fro') ** 2) / 2 )
                    loss.backward()

                # self.superresolution_optimizers[dataset.name].step()
                superresolution_optimizer.step()
                # self.superresolution_schedulers[dataset.name].step()

                if verbose > 4 and (iteration % 5 == 0):
                    pass
                    # print(f'NMF component: {((X @ MTM - YM) * X).sum().item() / 2 + Ynorm / 2}')
                    # print(f'Superresolution component: {((BTB @ X - BTX_B) * X).sum().item() / 2 + X_Bnorm / 2}')
                    # print(f'Total loss (recomputed): {((X @ MTM - YM + BTB @ X - BTX_B) * X).sum().item() / 2 + (Ynorm + X_Bnorm) / 2}')
                    # print(f"NMF component: {(torch.linalg.norm(Y.T - M @ X.T, ord='fro').item() ** 2)/ (sigma_yx ** 2)}")
                    # print(f"Superresolution component: {torch.linalg.norm(X_B - B @ X, ord='fro').item() ** 2}")
                    # print(f"Total loss (recomputed): {(torch.linalg.norm(Y.T - M @ X.T, ord='fro').item() ** 2)/ (sigma_yx ** 2) + torch.linalg.norm(X_B - B @ X, ord='fro').item() ** 2}")
                    # print(f"Total loss: {loss}")

                # gradient = quadratic_term_gradient - linear_term_gradient
                # X = X.sub(gradient, alpha=self.superresolution_lr)
                # X = torch.clip(X, min=1e-10)
                with torch.no_grad():
                    X.clamp_(min=1e-10)

                # del quadratic_term_gradient

                return loss

            progress_bar = trange(n_epochs, leave=True, disable=(verbose < 5), miniters=10000)
            for epoch in progress_bar:
                X_prev = X.clone().detach()
                if update_alg == "mu":
                    pass
                elif update_alg == "gd":
                    loss = gradient_update(X, iteration=epoch)

                dX = torch.abs((X_prev - X) / torch.linalg.norm(X, dim=1, ord=1, keepdim=True)).max().item()

                # if epoch % decay_period:
                #     tol /= 10

                do_stop = dX < tol
                description = (
                    f"Updating weights hierarchically: loss = {loss:.1e} "
                    f"%δloss = {(loss_prev - loss) / loss:.1e} "
                    f"%δX = {dX:.1e}"
                )

                progress_bar.set_description(description)

                loss_prev = loss
                if do_stop:
                    break

            progress_bar.close()
            self.embedding_optimizer.embedding_state[dataset.name][:] = X.clone().detach()

            final_losses[dataset_index] = loss.cpu().detach().numpy()

            # Delete dangling reference
            # TODO: can probably delete all this?
            del superresolution_optimizer
            # del self.superresolution_optimizers[dataset.name]

            del sigma_yx
            del Y
            del X
            del X_B
            del B
            del M
            del MTM
            del BTB
            del YM
            del BTX_B

        return final_losses

    def _reload_state(self):
        """Reload Popari state using results from saved datasets.

        Opposite of `synchronize_datasets`. Should be used rarely, e.g. when loading a trained
        model from memory.

        """
        for dataset_index, dataset in enumerate(self.datasets):
            self.parameter_optimizer.metagene_state[dataset.name][:] = torch.from_numpy(
                dataset.uns["M"][dataset.name],
            ).to(**self.initial_context)
            self.embedding_optimizer.embedding_state[dataset.name][:] = torch.from_numpy(dataset.obsm["X"]).to(
                **self.initial_context,
            )
            # self.parameter_optimizer.sigma_yxs[dataset_index] = dataset.uns["sigma_yx"] TODO: sigma_yx doesn't seem to be saved correctly due to issues with `merge_anndata`

            with torch.no_grad():
                self.parameter_optimizer.spatial_affinity_state[dataset.name][:] = torch.from_numpy(
                    dataset.uns["Sigma_x_inv"][dataset.name],
                ).to(**self.initial_context)

        if self.parameter_optimizer.metagene_mode == "differential":
            self.parameter_optimizer.metagene_state.reaverage()

        if self.parameter_optimizer.spatial_affinity_mode == "differential lookup":
            self.parameter_optimizer.spatial_affinity_state.reaverage()

        self.parameter_optimizer.update_sigma_yx()  # TODO: sigma_yx doesn't seem to be saved correctly due to issues with `merge_anndata`; try that instead of this

    def synchronize_datasets(self):
        """Synchronize datasets with learned view parameters and embeddings."""
        for dataset_index, dataset in enumerate(self.datasets):
            dataset.uns["M"][dataset.name] = (
                self.parameter_optimizer.metagene_state[dataset.name].cpu().detach().numpy()
            )
            dataset.obsm["X"] = self.embedding_optimizer.embedding_state[dataset.name].cpu().detach().numpy()
            dataset.uns["sigma_yx"] = self.parameter_optimizer.sigma_yxs[dataset_index]
            with torch.no_grad():
                dataset.uns["Sigma_x_inv"][dataset.name][:] = (
                    self.parameter_optimizer.spatial_affinity_state[dataset.name].cpu().detach().numpy()
                )

            dataset.uns["losses"]["nll_embeddings"].append(self.embedding_optimizer.nll_embeddings())
            dataset.uns["losses"]["nll_spatial_affinities"].append(self.parameter_optimizer.nll_spatial_affinities())
            dataset.uns["losses"]["nll_metagenes"].append(self.parameter_optimizer.nll_metagenes())
            dataset.uns["losses"]["nll_sigma_yx"].append(self.parameter_optimizer.nll_sigma_yx())
            dataset.uns["losses"]["nll"].append(self.nll())

        if self.parameter_optimizer.metagene_mode == "differential":
            M_bar = {
                group_name: self.parameter_optimizer.metagene_state.M_bar[group_name].cpu().detach().numpy()
                for group_name in self.parameter_optimizer.metagene_groups
            }
            for dataset in self.datasets:
                dataset.uns["M_bar"] = M_bar

        if self.parameter_optimizer.spatial_affinity_mode == "differential lookup":
            spatial_affinity_bar = {
                group_name: self.parameter_optimizer.spatial_affinity_state.spatial_affinity_bar[group_name]
                .cpu()
                .detach()
                .numpy()
                for group_name in self.parameter_optimizer.spatial_affinity_groups
            }
            for dataset in self.datasets:
                dataset.uns["spatial_affinity_bar"] = spatial_affinity_bar

    def nll(self, use_spatial=False):
        """Compute overall negative log-likelihood for current model
        parameters."""

        with torch.no_grad():
            total_loss = torch.zeros(1, **self.context)
            if use_spatial:
                weighted_total_cells = 0
                for dataset in self.datasets:
                    E_adjacency_list = self.embedding_optimizer.adjacency_lists[dataset.name]
                    weighted_total_cells += sum(map(len, E_adjacency_list))

            for dataset_index, dataset in enumerate(self.datasets):
                sigma_yx = self.parameter_optimizer.sigma_yxs[dataset_index]
                Y = self.Ys[dataset_index].to(self.context["device"])
                X = self.embedding_optimizer.embedding_state[dataset.name].to(self.context["device"])
                M = self.parameter_optimizer.metagene_state[dataset.name].to(self.context["device"])
                prior_x_mode = self.parameter_optimizer.prior_x_modes[dataset_index]
                beta = self.betas[dataset_index]
                prior_x = self.parameter_optimizer.prior_xs[dataset_index]

                # Precomputing quantities
                MTM = M.T @ M / (sigma_yx**2)
                YM = Y.to(M.device) @ M / (sigma_yx**2)
                Ynorm = torch.square(Y).sum() / (sigma_yx**2)
                S = torch.linalg.norm(X, dim=1, ord=1, keepdim=True)

                Z = X / S
                N, G = Y.shape

                loss = ((X @ MTM) * X).sum() / 2 - (X * YM).sum() + Ynorm / 2

                logZ_i_Y = torch.full((N,), G / 2 * np.log(2 * np.pi * sigma_yx**2), **self.context)
                if not use_spatial:
                    logZ_i_X = torch.full((N,), 0, **self.context)
                    if (prior_x[0] != 0).all():
                        logZ_i_X += torch.full((N,), self.K * torch.log(prior_x[0]).item(), **self.context)
                    log_partition_function = (logZ_i_Y + logZ_i_X).sum()
                else:
                    adjacency_matrix = self.embedding_optimizer.adjacency_matrices[dataset.name].to(
                        self.context["device"],
                    )
                    Sigma_x_inv = self.parameter_optimizer.spatial_affinity_state[dataset.name].to(
                        self.context["device"],
                    )
                    nu = adjacency_matrix @ Z
                    eta = nu @ Sigma_x_inv
                    logZ_i_s = torch.full((N,), 0, **self.context)
                    if (prior_x[0] != 0).all():
                        logZ_i_s = torch.full(
                            (N,),
                            -self.K * torch.log(prior_x[0]).item()
                            + torch.log(factorial(self.K - 1, exact=True)).item(),
                            **self.context,
                        )

                    logZ_i_z = integrate_of_exponential_over_simplex(eta)
                    log_partition_function = (logZ_i_Y + logZ_i_z + logZ_i_s).sum()

                    if prior_x_mode == "exponential shared fixed":
                        loss += prior_x[0][0] * S.sum()
                    elif not prior_x_mode:
                        pass
                    else:
                        raise NotImplementedError

                    if Sigma_x_inv is not None:
                        loss += (eta).mul(Z).sum() / 2

                    spatial_affinity_bars = None
                    if self.parameter_optimizer.spatial_affinity_state.mode == "differential lookup":
                        spatial_affinity_bars = [
                            self.parameter_optimizer.spatial_affinity_state.spatial_affinity_bar[group_name].detach()
                            for group_name in self.parameter_optimizer.spatial_affinity_tags[dataset.name]
                        ]

                    regularization = torch.zeros(1, **self.context)
                    if spatial_affinity_bars is not None:
                        group_weighting = 1 / len(spatial_affinity_bars)
                        for group_Sigma_x_inv_bar in spatial_affinity_bars:
                            regularization += (
                                group_weighting
                                * self.parameter_optimizer.lambda_Sigma_bar
                                * (group_Sigma_x_inv_bar - Sigma_x_inv).pow(2).sum()
                                / 2
                            )

                    regularization += self.parameter_optimizer.lambda_Sigma_x_inv * Sigma_x_inv.pow(2).sum() / 2

                    regularization *= weighted_total_cells

                    loss += regularization.item()

                loss += log_partition_function

                differential_regularization_term = torch.zeros(1, **self.context)
                M_bar = None
                if self.parameter_optimizer.metagene_mode == "differential":
                    M_bar = [
                        self.parameter_optimizer.metagene_state.M_bar[group_name]
                        for group_name in self.parameter_optimizer.metagene_tags[dataset.name]
                    ]

                if self.parameter_optimizer.lambda_M > 0 and M_bar is not None:
                    differential_regularization_quadratic_factor = self.parameter_optimizer.lambda_M * torch.eye(
                        self.K,
                        **self.context,
                    )

                    differential_regularization_linear_term = torch.zeros_like(M, **self.context)
                    group_weighting = 1 / len(M_bar)
                    for group_M_bar in M_bar:
                        differential_regularization_linear_term += (
                            group_weighting * self.parameter_optimizer.lambda_M * group_M_bar
                        )

                    differential_regularization_term = (
                        M @ differential_regularization_quadratic_factor * M
                    ).sum() - 2 * (differential_regularization_linear_term * M).sum()
                    group_weighting = 1 / len(M_bar)
                    for group_M_bar in M_bar:
                        differential_regularization_term += (
                            group_weighting * self.parameter_optimizer.lambda_M * (group_M_bar * group_M_bar).sum()
                        )

                loss += differential_regularization_term.item()

                total_loss += loss

        return total_loss.cpu().numpy()


class Hierarchy:
    """Container for hierarchical views of Popari."""

    def __init__(
        self,
        base_view: HierarchicalView,
        downsampling_method: str = "grid",
        **hierarchical_view_kwargs,
    ):
        self.view_container = {0: base_view}
        if downsampling_method == "grid":
            self.downsampler = GridDownsampler()
        elif downsampling_method == "partition":
            self.downsampler = PartitionDownsampler()

        self.hierarchical_view_kwargs = hierarchical_view_kwargs

    def __setitem__(self, index: int, view: HierarchicalView):
        self.view_container[index] = view

    def __getitem__(self, index: int):
        return self.view_container[index]

    def construct(self, levels: int, downsample_rate: float, **kwargs):
        base_view = self[0]
        context = base_view.context
        previous_view = base_view
        original_names = [dataset.name for dataset in previous_view.datasets]
        for level in range(1, levels):
            print(f"{get_datetime()} Initializing hierarchy level {level}")
            previous_datasets = previous_view.datasets
            binned_datasets = []
            binned_Ys = []
            previous_Ys = previous_view.Ys

            effective_kwargs = kwargs.copy()
            for previous_Y, previous_dataset, original_name in zip(previous_Ys, previous_datasets, original_names):
                dataset_name, *_ = previous_dataset.name.split("_level_")
                binned_dataset_name = f"{dataset_name}_level_{level}"
                bin_assignments_key = f"bin_assignments_{binned_dataset_name}"
                binned_dataset, effective_kwargs = self.downsampler.downsample(
                    previous_dataset,
                    downsample_rate=downsample_rate,
                    bin_assignments_key=bin_assignments_key,
                    **effective_kwargs,
                )
                binned_dataset = PopariDataset(binned_dataset, binned_dataset_name)
                binned_dataset.compute_spatial_neighbors()

                print(
                    f"{get_datetime()} Downsized dataset from {len(previous_dataset)} to {len(binned_dataset)} spots.",
                )

                binned_datasets.append(binned_dataset)
                bin_assignments = convert_numpy_to_pytorch_sparse_coo(
                    csr_array(binned_dataset.obsm[f"bin_assignments_{binned_dataset_name}"]).tocoo(),
                    context=context,
                )

                binned_Y = bin_assignments @ previous_Y
                binned_Ys.append(binned_Y)

            level_view = HierarchicalView(
                binned_datasets,
                level=level,
                binned_Ys=binned_Ys,
                **self.hierarchical_view_kwargs,
            )
            previous_view.link(level_view)
            self[level] = level_view

            previous_view = level_view

    @classmethod
    def reconstruct(cls, reloaded_hierarchy: dict, **hierarchical_view_kwargs):
        """Reconstruct hierarchy object from dictionary of level binned
        datasets."""

        context = hierarchical_view_kwargs["context"]

        def reconstruct_level(level: int, datasets: Sequence[PopariDataset], previous_view: HierarchicalView | None):
            print(f"{get_datetime()} Reloading level {level}")
            if previous_view is not None:
                binned_Ys = []
                for dataset, previous_Y in zip(datasets, previous_view.Ys):
                    B = dataset.obsm[f"bin_assignments_{dataset.name}"]
                    dataset.obsm[f"bin_assignments_{dataset.name}"] = csr_array(B)
                    binned_Y = (
                        convert_numpy_to_pytorch_sparse_coo(
                            dataset.obsm[f"bin_assignments_{dataset.name}"],
                            context=context,
                        )
                        @ previous_Y
                    )
                    binned_Ys.append(binned_Y)
            else:
                binned_Ys = None

            level_view = HierarchicalView(
                datasets,
                level=level,
                binned_Ys=binned_Ys,
                **hierarchical_view_kwargs,
            )

            level_view._reload_state()

            if previous_view is not None:
                previous_view.link(level_view)

            return level_view

        base_view = reconstruct_level(0, reloaded_hierarchy[0], None)

        hierarchy = cls(base_view=base_view, **hierarchical_view_kwargs)
        previous_view = base_view

        for level in range(1, hierarchical_view_kwargs["hierarchical_levels"]):
            datasets = reloaded_hierarchy[level]
            level_view = reconstruct_level(level, datasets, previous_view)
            hierarchy[level] = level_view
            previous_view = level_view

        return hierarchy



from functools import partial, wraps
from typing import Optional, Sequence

import numpy as np
import scanpy as sc
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from matplotlib.colors import ListedColormap
from scipy.stats import pearsonr, spearmanr, wasserstein_distance

from popari._dataset_utils import (
    _cluster,
    _cluster_domains,
    _compute_ari_score,
    _compute_columnwise_autocorrelation,
    _compute_confusion_matrix,
    _compute_empirical_correlations,
    _compute_silhouette_score,
    _compute_spatial_gene_correlation,
    _evaluate_classification_task,
    _leiden,
    _metagene_gsea,
    _multigroup_heatmap,
    _pca,
    _preprocess_embeddings,
    _umap,
    for_model,
    setup_squarish_axes,
)
from popari.model import Popari
from popari.util import spatially_smooth_feature

preprocess_embeddings = for_model(_preprocess_embeddings)
cluster = for_model(_cluster)
leiden = for_model(_leiden)
pca = for_model(_pca)
umap = for_model(_umap)
compute_ari_scores = for_model(_compute_ari_score)
compute_silhouette_scores = for_model(_compute_silhouette_score)
evaluate_classification_task = for_model(_evaluate_classification_task)
compute_confusion_matrix = for_model(_compute_confusion_matrix)
compute_columnwise_autocorrelation = for_model(_compute_columnwise_autocorrelation)
compute_spatial_gene_correlation = for_model(_compute_spatial_gene_correlation)
cluster_domains = for_model(_cluster_domains)


# def leiden(
#     trained_model: Popari,
#     resolution: float = 1.0,
#     tolerance: float = 0.05,
#     **kwargs,
# ):
#     r"""Compute Leiden clustering for all datasets.
#
#     Args:
#         trained_model: the trained Popari model.
#         joint: if `True`, jointly cluster the spots
#         use_rep: the key in the ``.obsm`` dataframe to ue as input to the Leiden clustering algorithm.
#         resolution: the resolution to use for Leiden clustering. Higher values yield finer clusters..
#
#     """
#     cluster(
#         trained_model,
#         resolution=resolution,
#         tolerance=tolerance,
#         flavor="igraph",
#         n_iterations=2,
#         **kwargs,
#     )


def multigroup_heatmap(
    trained_model: Popari,
    title_font_size: Optional[int] = None,
    group_type: str = "metagene",
    axes: Optional[Sequence[Axes]] = None,
    key: Optional[str] = None,
    level=0,
    **heatmap_kwargs,
):
    r"""Plot 2D heatmap data across all datasets.

    Wrapper function to enable plotting of continuous 2D data across multiple replicates. Only
    one of ``obsm``, ``obsp`` or ``uns`` should be used.

    Args:
        trained_model: the trained Popari model.
        axes: A predefined set of matplotlib axes to plot on.
        obsm: the key in the ``.obsm`` dataframe to plot.
        obsp: the key in the ``.obsp`` dataframe to plot.
        uns: the key in the ``.uns`` dataframe to plot. Unstructured data must be 2D in shape.
        **heatmap_kwargs: arguments to pass to the `ax.imshow` call for each dataset

    """
    datasets = trained_model.hierarchy[level].datasets

    groups = trained_model.metagene_groups if group_type == "metagene" else trained_model.spatial_affinity_groups
    _multigroup_heatmap(datasets, title_font_size=title_font_size, groups=groups, axes=axes, key=key, **heatmap_kwargs)


def compute_empirical_correlations(
    trained_model: Popari,
    feature: str = "X",
    level=0,
    output: str = "empirical_correlation",
):
    """Compute the empirical spatial correlation for a feature set across all
    datasets.

    Args:
        trained_model: the trained Popari model.
        feature: key in `.obsm` of feature set for which spatial correlation should be computed.
        output: key in `.uns` where output correlation matrices should be stored.

    """

    datasets = trained_model.hierarchy[level].datasets

    scaling = trained_model.parameter_optimizer.spatial_affinity_state.scaling

    _compute_empirical_correlations(datasets, scaling, feature=feature, output=output)


def find_differential_genes(trained_model: Popari, level=0, top_gene_limit: int = 1):
    """Identify genes/features that distinguish differential metagenes within a
    group.

    This type of analysis is only valid for runs of Popari in which ``metagene_mode="differential"``
    was used.

    Args:
        trained_model: the trained Popari model.
        top_gene_limit: the number of top genes to mark as differential for each metagene in a dataset.

    Returns:
        The names of genes that are differentially expressed with respect to their group.

    """

    datasets = trained_model.hierarchy[level].datasets

    for dataset in datasets:
        if "M_bar" not in dataset.uns:
            raise ValueError("This model was not trained in differential metagene mode.")

    genes_of_interest = set()
    for dataset in datasets:
        for group_name in trained_model.metagene_tags[dataset.name]:
            image = dataset.uns["M"][dataset.name] - dataset.uns["M_bar"][group_name]
            top_genes_per_metagene = np.argpartition(np.abs(image), -top_gene_limit, axis=0)[-top_gene_limit:]
            dataset_top_genes = dataset.var_names[top_genes_per_metagene.flatten()]
            genes_of_interest.update(dataset_top_genes)

    return genes_of_interest


def plot_gene_activations(trained_model: Popari, gene_subset: Sequence[str], level=0):
    """Plot metagene activation heatmaps for target genes across all groups.

    This type of analysis is only valid for runs of Popari in which ``metagene_mode="differential"``
    was used.

    Args:
        trained_model: the trained Popari model.
        gene_subset: names of genes to plot for.

    Returns:
        The names of genes that are differentially expressed with respect to their group.

    """
    gene_indices = trained_model.datasets[0].var_names.get_indexer(gene_subset)
    images = np.zeros((len(gene_indices), trained_model.K, len(trained_model.metagene_groups)))
    for group_index, group_name in enumerate(trained_model.metagene_groups):
        M_bar_subset = trained_model.datasets[0].uns["M_bar"][group_name][gene_indices]
        images[:, :, group_index] = M_bar_subset

    fig, axes = setup_squarish_axes(len(gene_indices), figsize=(10, 10))
    for ax, image, gene in zip(axes.flat, images, gene_subset):
        aspect = 0.1
        im = ax.imshow(image, interpolation="nearest", aspect=aspect)
        ax.set_title(gene)
        colorbar = fig.colorbar(im, ax=ax, orientation="vertical")


def plot_gene_trajectories(
    trained_model: Popari,
    gene_subset: Sequence[str],
    covariate_values: Sequence[float],
    level=0,
    **subplots_kwargs,
):
    """Plot metagene activation lineplots for target genes across all groups.

    This type of analysis is only valid for runs of Popari in which ``metagene_mode="differential"``
    was used.

    Args:
        trained_model: the trained Popari model.
        gene_subset: names of genes to plot for.
        covariate_values: dependent vairable values against which gene trajectories will be plotted.

    Returns:
        The names of genes that are differentially expressed with respect to their group.

    """
    gene_indices = trained_model.datasets[0].var_names.get_indexer(gene_subset)
    images = np.zeros((len(gene_indices), trained_model.K, len(trained_model.metagene_groups)))
    for group_index, group_name in enumerate(trained_model.metagene_groups):
        M_bar_subset = trained_model.parameter_optimizer.metagene_state.M_bar[group_name][gene_indices]
        images[:, :, group_index] = M_bar_subset

    summed_weights = images.sum(axis=1)
    fig, axes = setup_squarish_axes(len(gene_indices), figsize=(10, 10))
    for ax, trend, gene in zip(axes.flat, summed_weights, gene_subset):
        aspect = 0.1
        r = np.corrcoef(covariate_values, y=trend)[0, 1]
        im = ax.plot(covariate_values, trend)
        ax.set_title(f"{gene}, R = {r:.2f}")


def normalized_affinity_trends(
    trained_model,
    timepoint_values: Sequence[float],
    time_unit="Days",
    normalize: bool = False,
    spatial_affinity_key: str = "Sigma_x_inv",
    n_best: int = 5,
    highlight_metric: str = "pearson",
    figsize: tuple = None,
    margin_size: float = 0.25,
    level=0,
):
    """Plot trends for every pair of affinities; highlight top trends.

    Args:
        trained_model: the trained Popari model.
        timepoint_values: x-values against which to plot trends
        time_unit: unit in which time is measured (used for x-axis label)

    """

    datasets = trained_model.hierarchy[level].datasets
    all_affinities = np.array([dataset.uns[spatial_affinity_key][dataset.name] for dataset in datasets])

    if normalize:
        for index in range(len(datasets), axes.size):
            prenormalization_affinity_std = np.std(all_affinities, axis=0, keepdims=True)
            prenormalization_timepoint_std = np.std(timepoint_values)
            all_affinities /= prenormalization_affinity_std
            timepoint_values /= prenormalization_timepoint_std

    timepoint_min = np.min(timepoint_values)
    timepoint_ptp = np.ptp(timepoint_values)
    timepoint_std = np.std(timepoint_values)

    affinity_min = np.min(all_affinities)
    affinity_ptp = np.ptp(all_affinities)
    affinity_std = np.std(all_affinities, axis=0, keepdims=True)

    pearson_correlations = {}
    variances = {}
    pearson_p_values = {}
    slopes = {}

    for i in range(trained_model.K):
        for j in range(i + 1):
            affinity_values = all_affinities[:, i, j]
            r, p_value = pearsonr(affinity_values, timepoint_values)
            variances[(i, j)] = np.var(affinity_values)
            pearson_correlations[(i, j)] = r
            pearson_p_values[(i, j)] = r
            slopes[(i, j)] = r * affinity_std[0, i, j] / timepoint_std

    if highlight_metric == "pearson":
        pairs, metric_values = zip(*pearson_correlations.items())
    elif highlight_metric == "variance":
        pairs, metric_values = zip(*variances.items())

    pairs = np.array(pairs)
    sorted_values = np.argsort(metric_values)
    best_values = sorted_values[-n_best:][::-1] if n_best > 0 else sorted_values[:-n_best]

    top_pairs = pairs[best_values].tolist()

    for dataset in datasets:
        dataset.uns["spatial_trends"] = {
            "top_pairs": top_pairs,
            "pearson_correlations": pearson_correlations,
            "variances": variances,
            "slopes": slopes,
        }

    return top_pairs, pearson_correlations, variances


def propagate_labels(trained_model, label_key: str, starting_level=None, smooth=False):
    """Propagate a label from the most binned layer to the least binned
    layer."""

    if starting_level is None:
        starting_level = trained_model.hierarchical_levels - 1

    for level in range(starting_level, 0, -1):
        view = trained_model.hierarchy[level]
        high_res_view = trained_model.hierarchy[level - 1]
        for dataset, high_res_dataset in zip(view.datasets, high_res_view.datasets):
            B = dataset.obsm[f"bin_assignments_{dataset.name}"]
            labels = dataset.obs[label_key]

            high_res_labels = np.array(labels) @ B.astype(int).toarray()
            if smooth:
                high_res_labels = spatially_smooth_feature(
                    high_res_labels,
                    high_res_dataset.obs["adjacency_list"],
                    max_smoothing_rounds=200,
                    smoothing_threshold=0.3,
                )

            high_res_dataset.obs[label_key] = list(high_res_labels)
            high_res_dataset.obs[label_key] = high_res_dataset.obs[label_key].astype("category")


def metagene_gsea(trained_model, metagene_index: int, level=0, **gsea_kwargs):
    """Run GSEA on metagenes from trained model."""

    first_dataset = trained_model.hierarchy[level].datasets[0]
    fig = _metagene_gsea(first_dataset, metagene_index, **gsea_kwargs)

    return fig
